# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QSpinBox,
    QVBoxLayout,
)

MAX_INT = 2147483647


class FindEventsDialog(QDialog):
    def __init__(self, parent, channels, default_stim):
        super().__init__(parent)
        self.setWindowTitle("Find Events")
        vbox = QVBoxLayout(self)
        grid = QGridLayout()

        grid.addWidget(QLabel("Stim channel:"), 0, 0)
        self.stimchan = QComboBox()
        self.stimchan.addItems(channels)
        self.stimchan.setCurrentIndex(default_stim)
        grid.addWidget(self.stimchan, 0, 1)

        grid.addWidget(QLabel("Consecutive"), 1, 0)
        self.consecutive = QComboBox()
        self.consecutive.addItems(["Increasing", "True", "False"])
        self.consecutive.setCurrentIndex(0)
        grid.addWidget(self.consecutive, 1, 1)

        grid.addWidget(QLabel("Initial event"), 2, 0)
        self.initial_event = QCheckBox()
        self.initial_event.setChecked(True)
        grid.addWidget(self.initial_event, 2, 1)

        grid.addWidget(QLabel("Interpret as uint16"), 3, 0)
        self.uint_cast = QCheckBox()
        self.uint_cast.setChecked(True)
        grid.addWidget(self.uint_cast, 3, 1)

        grid.addWidget(QLabel("Minimum duration:"), 4, 0)
        self.minduredit = QSpinBox()
        self.minduredit.setMaximum(MAX_INT)
        grid.addWidget(self.minduredit, 4, 1)

        grid.addWidget(QLabel("Shortest event:"), 5, 0)
        self.shortesteventedit = QSpinBox()
        self.shortesteventedit.setValue(2)
        self.shortesteventedit.setMaximum(MAX_INT)
        grid.addWidget(self.shortesteventedit, 5, 1)

        vbox.addLayout(grid)
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(buttonbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)



import re

from mne import create_info
from mne.io import RawArray
from numpy import atleast_2d
from scipy.io import loadmat


def read_raw_mat(fname, variable, fs, transpose=False, *args, **kwargs):
    """Read raw data from MAT file.

    Parameters
    ----------
    fname : str
        File name to load.
    variable : str
        Name of the variable to use. If nested within a struct, separate all names with
        dots. For example, "y.X" corresponds to a variable "X" contained in a struct "y".
    fs : float
        Sampling frequency (in Hz).
    transpose : bool
        Whether to transpose the data.
    """
    mat = loadmat(fname, simplify_cells=True)
    data = atleast_2d(_get_dict_value(mat, variable.split(".")))
    if transpose:
        data = data.T
    info = create_info(data.shape[0], fs, "eeg")
    raw = RawArray(data, info)
    raw._filenames = [fname]
    return raw


def parse_mat(fname):
    """Remove dunder variables from dict returned by scipy.io.loadmat()."""
    mat = loadmat(fname, simplify_cells=True)
    return {k: v for k, v in mat.items() if not k.startswith("__") and not k.endswith("__")}


def _get_dict_value(d, keys):
    """Get dictionary value from nested dictionary keys."""
    if isinstance(keys, str):  # no nesting
        return d[keys]
    value = d
    for key in keys:
        if match := re.search(r"\[(\d+)\]", key):  # list element
            value = value[int(match.group(1))]
        else:  # dict element
            value = value[key]
    return value



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QListWidget,
    QPushButton,
    QVBoxLayout,
)


class AppendDialog(QDialog):
    def __init__(self, parent, compatibles, title="Append data"):
        super().__init__(parent)
        self.setWindowTitle(title)

        vbox = QVBoxLayout(self)
        grid = QGridLayout()

        grid.addWidget(QLabel("Source"), 0, 0, Qt.AlignCenter)
        grid.addWidget(QLabel("Destination"), 0, 2, Qt.AlignCenter)

        self.source = QListWidget(self)
        self.source.setAcceptDrops(True)
        self.source.setDragEnabled(True)
        self.source.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.source.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.source.insertItems(0, [d["name"] for d in compatibles])
        grid.addWidget(self.source, 1, 0)

        self.move_button = QPushButton("→")
        self.move_button.setEnabled(False)
        grid.addWidget(self.move_button, 1, 1, Qt.AlignHCenter)

        self.destination = QListWidget(self)
        self.destination.setAcceptDrops(True)
        self.destination.setDragEnabled(True)
        self.destination.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.destination.setDefaultDropAction(Qt.DropAction.MoveAction)
        grid.addWidget(self.destination, 1, 2)
        vbox.addLayout(grid)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        vbox.addWidget(self.buttonbox)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)
        self.destination.model().rowsInserted.connect(self.toggle_ok_button)
        self.destination.model().rowsRemoved.connect(self.toggle_ok_button)
        self.source.itemSelectionChanged.connect(self.toggle_move_source)
        self.destination.itemSelectionChanged.connect(self.toggle_move_destination)
        self.move_button.clicked.connect(self.move)
        self.toggle_ok_button()
        self.toggle_move_source()
        self.toggle_move_destination()

    @property
    def names(self):
        names = []
        for it in range(self.destination.count()):
            names.append(self.destination.item(it).text())
        return names

    @Slot()
    def toggle_ok_button(self):
        """Toggle OK button."""
        if self.destination.count() > 0:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)

    @Slot()
    def toggle_move_source(self):
        if self.source.selectedItems():
            self.move_button.setEnabled(True)
            self.move_button.setText("→")
            self.destination.clearSelection()
        elif not self.destination.selectedItems():
            self.move_button.setEnabled(False)

    @Slot()
    def toggle_move_destination(self):
        if self.destination.selectedItems():
            self.move_button.setEnabled(True)
            self.move_button.setText("←")
            self.source.clearSelection()
        elif not self.source.selectedItems():
            self.move_button.setEnabled(False)

    @Slot()
    def move(self):
        if self.source.selectedItems():
            for item in self.source.selectedItems():
                self.destination.addItem(item.text())
                self.source.takeItem(self.source.row(item))
        elif self.destination.selectedItems():
            for item in self.destination.selectedItems():
                self.source.addItem(item.text())
                self.destination.takeItem(self.destination.row(item))



# © MNELAB developers
#
# License: BSD (3-clause)
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QGridLayout,
    QLabel,
    QListWidget,
    QVBoxLayout,
)

from .utils import select_all


class ERDSDialog(QDialog):
    def __init__(self, parent, t_range, f_range):
        super().__init__(parent)
        self.setWindowTitle("ERDS maps")
        vbox = QVBoxLayout(self)
        grid = QGridLayout()

        grid.addWidget(QLabel("Frequency range:"), 0, 0)
        self._f1 = QDoubleSpinBox()
        self._f1.setRange(*f_range)
        self._f1.setValue(f_range[0])
        self._f1.setDecimals(1)
        self._f1.setSuffix(" Hz")
        grid.addWidget(self._f1, 0, 1)

        self._f2 = QDoubleSpinBox()
        self._f2.setRange(*f_range)
        self._f2.setValue(f_range[1])
        self._f2.setDecimals(1)
        self._f2.setSuffix(" Hz")
        grid.addWidget(self._f2, 0, 2)

        grid.addWidget(QLabel("Step size:"), 1, 0)
        self._step = QDoubleSpinBox()
        self._step.setRange(0.1, 5)
        self._step.setValue(1)
        self._step.setDecimals(1)
        self._step.setSingleStep(0.1)
        self._step.setSuffix(" Hz")
        grid.addWidget(self._step, 1, 1)

        grid.addWidget(QLabel("Time range:"), 2, 0)
        self._t1 = QDoubleSpinBox()
        self._t1.setRange(*t_range)
        self._t1.setValue(t_range[0])
        self._t1.setDecimals(1)
        self._t1.setSingleStep(0.1)
        self._t1.setSuffix(" s")
        grid.addWidget(self._t1, 2, 1)

        self._t2 = QDoubleSpinBox()
        self._t2.setRange(*t_range)
        self._t2.setValue(t_range[1])
        self._t2.setDecimals(1)
        self._t2.setSingleStep(0.1)
        self._t2.setSuffix(" s")
        grid.addWidget(self._t2, 2, 2)

        grid.addWidget(QLabel("Baseline:"), 3, 0)
        self._b1 = QDoubleSpinBox()
        self._b1.setRange(*t_range)
        self._b1.setValue(t_range[0])
        self._b1.setDecimals(1)
        self._b1.setSingleStep(0.1)
        self._b1.setSuffix(" s")
        grid.addWidget(self._b1, 3, 1)

        self._b2 = QDoubleSpinBox()
        self._b2.setRange(*t_range)
        self._b2.setValue(0)
        self._b2.setDecimals(1)
        self._b2.setSingleStep(0.1)
        self._b2.setSuffix(" s")
        grid.addWidget(self._b2, 3, 2)

        self.significance_mask = QCheckBox("Significance level:")
        self.significance_mask.setChecked(False)
        self.alpha = QDoubleSpinBox()
        self.alpha.setMinimum(0)
        self.alpha.setValue(0.05)
        self.alpha.setDecimals(2)
        self.alpha.setSingleStep(0.01)
        grid.addWidget(self.significance_mask, 4, 0)
        grid.addWidget(self.alpha, 4, 1)
        self.significance_mask.toggled.connect(self.toggle_alpha)
        self.toggle_alpha()

        vbox.addLayout(grid)
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(buttonbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    @property
    def f1(self):
        return self._f1.value()

    @property
    def f2(self):
        return self._f2.value()

    @property
    def step(self):
        return self._step.value()

    @property
    def t1(self):
        return self._t1.value()

    @property
    def t2(self):
        return self._t2.value()

    @property
    def b1(self):
        return self._b1.value()

    @property
    def b2(self):
        return self._b2.value()

    @Slot()
    def toggle_alpha(self):
        self.alpha.setEnabled(self.significance_mask.isChecked())


class ERDSTopomapsDialog(QDialog):
    def __init__(self, parent, t_range, f_range, events):
        super().__init__(parent)
        self.setWindowTitle("ERDS topomaps")
        vbox = QVBoxLayout(self)
        grid = QGridLayout()

        grid.addWidget(QLabel("Frequency range:"), 0, 0)
        self._f1 = QDoubleSpinBox()
        self._f1.setRange(*f_range)
        self._f1.setValue(f_range[0])
        self._f1.setDecimals(1)
        self._f1.setSuffix(" Hz")
        grid.addWidget(self._f1, 0, 1)

        self._f2 = QDoubleSpinBox()
        self._f2.setRange(*f_range)
        self._f2.setValue(f_range[1])
        self._f2.setDecimals(1)
        self._f2.setSuffix(" Hz")
        grid.addWidget(self._f2, 0, 2)

        grid.addWidget(QLabel("Time range:"), 1, 0)
        self._t1 = QDoubleSpinBox()
        self._t1.setRange(*t_range)
        self._t1.setValue(t_range[0])
        self._t1.setDecimals(1)
        self._t1.setSingleStep(0.1)
        self._t1.setSuffix(" s")
        grid.addWidget(self._t1, 1, 1)

        self._t2 = QDoubleSpinBox()
        self._t2.setRange(*t_range)
        self._t2.setValue(t_range[1])
        self._t2.setDecimals(1)
        self._t2.setSingleStep(0.1)
        self._t2.setSuffix(" s")
        grid.addWidget(self._t2, 1, 2)

        grid.addWidget(QLabel("Baseline:"), 2, 0)
        self._b1 = QDoubleSpinBox()
        self._b1.setRange(*t_range)
        self._b1.setValue(t_range[0])
        self._b1.setDecimals(1)
        self._b1.setSingleStep(0.1)
        self._b1.setSuffix(" s")
        grid.addWidget(self._b1, 2, 1)

        self._b2 = QDoubleSpinBox()
        self._b2.setRange(*t_range)
        self._b2.setValue(0)
        self._b2.setDecimals(1)
        self._b2.setSingleStep(0.1)
        self._b2.setSuffix(" s")
        grid.addWidget(self._b2, 2, 2)

        label = QLabel("Events:")
        label.setAlignment(Qt.AlignTop)
        grid.addWidget(label, 3, 0)
        self.events = QListWidget()
        self.events.insertItems(0, events)
        self.events.setSelectionMode(QListWidget.ExtendedSelection)
        self.events.setMaximumHeight(self.events.sizeHintForRow(0) * 5.5)
        select_all(self.events)
        grid.addWidget(self.events, 3, 1, 1, 2)

        grid.addWidget(QLabel("Frequency step:"), 4, 0)
        self._step = QDoubleSpinBox()
        self._step.setRange(0.1, 5)
        self._step.setValue(1)
        self._step.setDecimals(1)
        self._step.setSingleStep(0.1)
        self._step.setSuffix(" Hz")
        grid.addWidget(self._step, 4, 1)

        vbox.addLayout(grid)
        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        self.events.itemSelectionChanged.connect(self.toggle_ok)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    @Slot()
    def toggle_ok(self):
        enable = bool(self.events.selectedItems())
        self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(enable)

    @property
    def f1(self):
        return self._f1.value()

    @property
    def f2(self):
        return self._f2.value()

    @property
    def step(self):
        return self._step.value()

    @property
    def t1(self):
        return self._t1.value()

    @property
    def t2(self):
        return self._t2.value()

    @property
    def b1(self):
        return self._b1.value()

    @property
    def b2(self):
        return self._b2.value()



# © MNELAB developers
#
# License: BSD (3-clause)

import re
from pathlib import Path

import numpy as np


def count_locations(info):
    locs = np.array([ch["loc"][:3] for ch in info["chs"]])
    valid_locs = np.any(~np.isclose(locs, 0) & np.isfinite(locs), axis=1)
    return valid_locs.sum()


def image_path(fname):
    """Return absolute path to image fname."""
    root = Path(__file__).parent.parent
    return str((root / "images" / Path(fname)).resolve())


def interface_style():
    """Return current platform interface style (light or dark)."""
    try:  # currently only works on macOS
        from Foundation import NSUserDefaults as NSUD
    except ImportError:
        return None
    style = NSUD.standardUserDefaults().stringForKey_("AppleInterfaceStyle")
    if style == "Dark":
        return "dark"
    else:
        return "light"


def natural_sort(lst):
    """Sort a list in natural order."""
    def key(s):
        return [int(t) if t.isdigit() else t.lower() for t in re.split(r'(\d+)', str(s))]
    return sorted(lst, key=key)



import struct
import xml.etree.ElementTree as ETree
from collections import defaultdict

import mne
import numpy as np
import scipy.signal
from mne.io.pick import get_channel_type_constants
from pyxdf import load_xdf
from pyxdf.pyxdf import _read_varlen_int, open_xdf


def _resample_streams(streams, stream_ids, fs_new):
    """
    Resample multiple XDF streams to a given frequency.

    Parameters
    ----------
    streams : dict
        A dictionary mapping stream IDs to XDF streams.
    stream_ids : list[int]
        The IDs of the desired streams.
    fs_new : float
        Resampling target frequency in Hz.

    Returns
    -------
    all_time_series : np.ndarray
        Array of shape (n_samples, n_channels) containing raw data. Time intervals where a
        stream has no data contain `np.nan`.
    first_time : float
        Time of the very first sample in seconds.
    """
    start_times = []
    end_times = []
    n_total_chans = 0
    for stream_id in stream_ids:
        start_times.append(streams[stream_id]["time_stamps"][0])
        end_times.append(streams[stream_id]["time_stamps"][-1])
        n_total_chans += int(streams[stream_id]["info"]["channel_count"][0])
    first_time = min(start_times)
    last_time = max(end_times)

    n_samples = int(np.ceil((last_time - first_time) * fs_new))
    all_time_series = np.full((n_samples, n_total_chans), np.nan)

    col_start = 0
    for stream_id in stream_ids:
        start_time = streams[stream_id]["time_stamps"][0]
        end_time = streams[stream_id]["time_stamps"][-1]
        len_new = int(np.ceil((end_time - start_time) * fs_new))

        x_old = streams[stream_id]["time_series"]
        x_new = scipy.signal.resample(x_old, len_new, axis=0)

        row_start = int(np.floor((streams[stream_id]["time_stamps"][0] - first_time) * fs_new))  # noqa: E501
        row_end = row_start + x_new.shape[0]
        col_end = col_start + x_new.shape[1]
        all_time_series[row_start:row_end, col_start:col_end] = x_new

        col_start += x_new.shape[1]

    return all_time_series, first_time


def read_raw_xdf(
    fname,
    stream_ids,
    prefix_markers=False,
    fs_new=None,
    *args,
    **kwargs,
):
    """Read XDF file.

    Parameters
    ----------
    fname : str
        Name of the XDF file.
    stream_ids : list[int]
        IDs of the streams to load.
    prefix_markers : bool
        Whether or not to prefix markers with their corresponding stream ID.
    fs_new : float | None
        Resampling target frequency in Hz. If only one stream_id is given, this can be
        `None`, in which case no resampling is performed.

    Returns
    -------
    raw : mne.io.Raw
        XDF file data.
    """
    if len(stream_ids) > 1 and fs_new is None:
        raise ValueError("Parameter 'fs_new' required when reading multiple streams.")

    streams, _ = load_xdf(fname)
    streams = {stream["info"]["stream_id"]: stream for stream in streams}

    labels_all, types_all, units_all = [], [], []
    for stream_id in stream_ids:
        stream = streams[stream_id]

        n_chans = int(stream["info"]["channel_count"][0])
        labels, types, units = [], [], []
        try:
            for ch in stream["info"]["desc"][0]["channels"][0]["channel"]:
                labels.append(str(ch["label"][0]))
                if ch["type"] and ch["type"][0].lower() in get_channel_type_constants(include_defaults=True):  # noqa: E501
                    types.append(ch["type"][0].lower())
                else:
                    types.append("misc")
                units.append(ch["unit"][0] if ch["unit"] else "NA")
        except (TypeError, IndexError):  # no channel labels found
            pass
        if not labels:
            labels = [f"{stream['info']['name'][0]}_{n}" for n in range(n_chans)]
        if not units:
            units = ["NA" for _ in range(n_chans)]
        if not types:
            types = ["misc" for _ in range(n_chans)]
        labels_all.extend(labels)
        types_all.extend(types)
        units_all.extend(units)

    if fs_new is not None:
        all_time_series, first_time = _resample_streams(streams, stream_ids, fs_new)
        fs = fs_new
    else:  # only possible if a single stream was selected
        all_time_series = streams[stream_ids[0]]["time_series"]
        first_time = streams[stream_ids[0]]["time_stamps"][0]
        fs = float(np.array(stream["info"]["effective_srate"]).item())

    info = mne.create_info(ch_names=labels_all, sfreq=fs, ch_types=types_all)

    scale = np.array([1e-6 if u in ("microvolt", "microvolts") else 1 for u in units_all])
    all_time_series_scaled = (all_time_series * scale).T

    raw = mne.io.RawArray(all_time_series_scaled, info)
    raw._filenames = [fname]

    # convert marker streams to annotations
    for stream_id, stream in streams.items():
        srate = float(stream["info"]["nominal_srate"][0])
        n_chans = int(stream["info"]["channel_count"][0])
        # marker streams with regular srate or more than 1 channel are not supported yet
        if srate != 0 or n_chans > 1:
            continue
        onsets = stream["time_stamps"] - first_time
        prefix = f"{stream_id}-" if prefix_markers else ""
        descriptions = [f"{prefix}{item}" for sub in stream["time_series"] for item in sub]
        raw.annotations.append(onsets, [0] * len(onsets), descriptions)

    return raw


def get_xml(fname):
    """Get XML stream headers and footers from all streams.

    Parameters
    ----------
    fname : str
        Name of the XDF file.

    Returns
    -------
    xml : dict
        XML stream headers and footers.
    """
    with open_xdf(fname) as f:
        xml = defaultdict(dict)
        while True:
            try:
                nbytes = _read_varlen_int(f)
            except EOFError:
                return xml
            tag = struct.unpack('<H', f.read(2))[0]
            if tag in [2, 3, 4, 6]:
                stream_id = struct.unpack("<I", f.read(4))[0]
                if tag in [2, 6]:  # parse StreamHeader/StreamFooter chunk
                    string = f.read(nbytes - 6).decode()
                    xml[stream_id][tag] = ETree.fromstring(string)
                else:  # skip remaining chunk contents
                    f.seek(nbytes - 6, 1)
            else:
                f.seek(nbytes - 2, 1)  # skip remaining chunk contents


def list_chunks(fname):
    """List all chunks contained in an XDF file.

    Listing chunks summarizes the content of the XDF file. Because this function does not
    attempt to parse the data, this also works for corrupted files.

    Parameters
    ----------
    fname : str
        Name of the XDF file.

    Returns
    -------
    chunks : list
        List of dicts containing a short summary for each chunk.
    """
    with open_xdf(fname) as f:
        chunks = []
        while True:
            try:
                nbytes = _read_varlen_int(f)
            except EOFError:
                return chunks
            chunk = {"nbytes": nbytes}
            tag = struct.unpack('<H', f.read(2))[0]
            chunk["tag"] = tag
            if tag == 1:
                chunk["content"] = f.read(nbytes - 2).decode()
            elif tag == 5:
                chunk["content"] = ("0x43 0xA5 0x46 0xDC 0xCB 0xF5 0x41 0x0F 0xB3 0x0E "
                                    "0xD5 0x46 0x73 0x83 0xCB 0xE4")
                f.seek(chunk["nbytes"] - 2, 1)  # skip remaining chunk contents
            elif tag in [2, 6]:  # XML
                chunk["stream_id"] = struct.unpack("<I", f.read(4))[0]
                chunk["content"] = f.read(chunk["nbytes"] - 6).decode().replace("\t", "  ")
            elif tag == 4:
                chunk["stream_id"] = struct.unpack("<I", f.read(4))[0]
                collection_time = struct.unpack("<d", f.read(8))[0]
                offset_value = struct.unpack("<d", f.read(8))[0]
                chunk["content"] = (f"Collection time: {collection_time}\n"
                                    f"Offset value: {offset_value}")
            elif tag == 3:
                chunk["stream_id"] = struct.unpack("<I", f.read(4))[0]
                remainder = chunk["nbytes"] - 6
                chunk["content"] = f"<BINARY DATA ({remainder} Bytes)>"
                f.seek(remainder, 1)  # skip remaining chunk contents
            else:
                f.seek(chunk["nbytes"] - 2, 1)  # skip remaining chunk contents
            chunks.append(chunk)



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import QFont, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QPlainTextEdit,
    QTableView,
    QVBoxLayout,
)


def _add_item(item):
    tmp = QStandardItem()
    tmp.setData(item, Qt.DisplayRole)
    return tmp


class XDFChunksDialog(QDialog):
    def __init__(self, parent, chunks, fname):
        super().__init__(parent)
        self.setWindowTitle(f"XDF Chunks ({fname})")

        self.chunks = chunks

        TAGS = {1: "FileHeader", 2: "StreamHeader", 3: "Samples", 4: "ClockOffset",
                5: "Boundary", 6: "StreamFooter"}

        self.model = QStandardItemModel()
        self.model.setHorizontalHeaderLabels(["#", "Bytes", "Tag", "Stream ID"])

        for i, chunk in enumerate(chunks, start=1):
            row = []
            row.append(_add_item(i))
            row.append(_add_item(chunk["nbytes"]))
            row.append(_add_item(f"{chunk['tag']} ({TAGS[chunk['tag']]})"))
            row.append(_add_item(chunk.get("stream_id", "")))
            self.model.appendRow(row)

        self.view = QTableView()
        self.view.setModel(self.model)
        self.view.verticalHeader().setVisible(False)
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.setShowGrid(False)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSortingEnabled(True)
        self.view.sortByColumn(0, Qt.AscendingOrder)
        self.view.setEditTriggers(QTableView.NoEditTriggers)
        self.view.selectRow(0)
        self.view.setFixedWidth(450)

        self.details = QPlainTextEdit("")
        self.details.setFixedWidth(450)
        self.details.setReadOnly(True)
        self.details.setTabStopDistance(30)
        font = QFont()
        font.setFamily("monospace")
        font.setStyleHint(QFont.Monospace)
        self.details.setFont(font)
        self.details.setLineWrapMode(QPlainTextEdit.NoWrap)

        hbox = QHBoxLayout()
        hbox.addWidget(self.view)
        hbox.addWidget(self.details)

        vbox = QVBoxLayout(self)
        vbox.addLayout(hbox)
        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Close)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.rejected.connect(self.reject)

        self.view.clicked.connect(self._update_details)
        self._update_details()

        self.setFixedSize(980, 650)
        self.view.setColumnWidth(0, 70)
        self.view.setColumnWidth(1, 80)
        self.view.setColumnWidth(2, 150)
        self.view.setColumnWidth(3, 70)

    @Slot()
    def _update_details(self):
        selection = self.view.selectionModel()
        if selection.hasSelection():
            n = int(selection.selectedIndexes()[0].data())
            self.details.setPlainText(self.chunks[n - 1].get("content", ""))



# © MNELAB developers
#
# License: BSD (3-clause)

from importlib import import_module

required = ["mne", "numpy", "scipy", "matplotlib", "PySide6", "pyxdf"]
optional = ["sklearn", "picard", "pyedflib", "pybv"]

have = {}
for package in required + optional:
    try:
        mod = import_module(package)
    except ModuleNotFoundError:
        have[package] = False
    else:  # module successfully imported
        try:
            version = mod.__version__
        except AttributeError:
            version = "unknown"
        have[package] = version



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QSpinBox,
    QVBoxLayout,
)


class RunICADialog(QDialog):
    def __init__(self, parent, nchan, methods):
        super().__init__(parent)
        self.setWindowTitle("Run ICA")

        vbox = QVBoxLayout(self)
        grid = QGridLayout()
        grid.addWidget(QLabel("Method:"), 0, 0)
        self.method = QComboBox()
        self.method.addItems(methods)
        self.method.setCurrentIndex(0)
        self.method.currentIndexChanged.connect(self.toggle_options)
        grid.addWidget(self.method, 0, 1)

        self.extended_label = QLabel("Extended:")
        grid.addWidget(self.extended_label, 1, 0)
        self.extended = QCheckBox()
        self.extended.setChecked(True)
        grid.addWidget(self.extended, 1, 1)

        self.ortho_label = QLabel("Orthogonal:")
        grid.addWidget(self.ortho_label, 2, 0)
        self.ortho = QCheckBox()
        self.ortho.setChecked(False)
        grid.addWidget(self.ortho, 2, 1)
        if "Picard" not in methods:
            self.ortho_label.hide()
            self.ortho.hide()

        grid.addWidget(QLabel("Number of components:"), 3, 0)
        self.n_components = QSpinBox()
        self.n_components.setRange(0, nchan)
        self.n_components.setValue(nchan)
        self.n_components.setAlignment(Qt.AlignRight)
        grid.addWidget(self.n_components, 3, 1)

        grid.addWidget(QLabel("Exclude bad segments:"), 4, 0)
        self.exclude_bad_segments = QCheckBox()
        self.exclude_bad_segments.setChecked(True)
        grid.addWidget(self.exclude_bad_segments, 4, 1)

        vbox.addLayout(grid)

        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(buttonbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

        self.toggle_options()

    @Slot()
    def toggle_options(self):
        """Toggle extended options."""
        if self.method.currentText() == "Picard":  # enable extended and ortho
            self.extended_label.setEnabled(True)
            self.extended.setEnabled(True)
            self.ortho_label.setEnabled(True)
            self.ortho.setEnabled(True)
        elif self.method.currentText() == "Infomax":  # enable extended
            self.extended_label.setEnabled(True)
            self.extended.setEnabled(True)
            self.ortho_label.setEnabled(False)
            self.ortho.setChecked(False)
            self.ortho.setEnabled(False)
        else:
            self.extended_label.setEnabled(False)
            self.extended.setChecked(False)
            self.extended.setEnabled(False)
            self.ortho_label.setEnabled(False)
            self.ortho.setChecked(False)
            self.ortho.setEnabled(False)



# © MNELAB developers
#
# License: BSD (3-clause)

from .annotations import AnnotationsDialog
from .append import AppendDialog
from .calc import CalcDialog
from .channel_properties import ChannelPropertiesDialog
from .crop import CropDialog
from .drop_bad_epochs import DropBadEpochsDialog
from .epoch import EpochDialog
from .erds import ERDSDialog, ERDSTopomapsDialog
from .error_message import ErrorMessageBox
from .events import EventsDialog
from .filter import FilterDialog
from .find_events import FindEventsDialog
from .history import HistoryDialog
from .interpolate_bads import InterpolateBadsDialog
from .mat import MatDialog
from .montage import MontageDialog
from .pick_channels import PickChannelsDialog
from .plot_evoked import PlotEvokedComparisonDialog, PlotEvokedDialog, PlotEvokedTopomaps
from .reference import ReferenceDialog
from .rename_channels import RenameChannelsDialog
from .run_ica import RunICADialog
from .xdf_chunks import XDFChunksDialog
from .xdf_metadata import XDFMetadataDialog
from .xdf_streams import XDFStreamsDialog



# © MNELAB developers
#
# License: BSD (3-clause)

from functools import partial
from pathlib import Path

import mne

from .mat import read_raw_mat
from .xdf import read_raw_xdf


def _read_unsupported(fname, **kwargs):
    ext = "".join(Path(fname).suffixes)
    msg = f"Unsupported file type ({ext})."
    suggest = kwargs.get("suggest")
    if suggest is not None:
        msg += f" Try reading a {suggest} file instead."
    raise ValueError(msg)


# supported read file formats
supported = {
    ".edf": mne.io.read_raw_edf,
    ".bdf": mne.io.read_raw_bdf,
    ".gdf": mne.io.read_raw_gdf,
    ".vhdr": mne.io.read_raw_brainvision,
    ".fif": mne.io.read_raw_fif,
    ".fif.gz": mne.io.read_raw_fif,
    ".set": mne.io.read_raw_eeglab,
    ".cnt": mne.io.read_raw_cnt,
    ".mff": mne.io.read_raw_egi,
    ".nxe": mne.io.read_raw_eximia,
    ".hdr": mne.io.read_raw_nirx,
    ".snirf": mne.io.read_raw_snirf,
    ".xdf": read_raw_xdf,
    ".xdfz": read_raw_xdf,
    ".xdf.gz": read_raw_xdf,
    ".mat": read_raw_mat,
}

# known but unsupported file formats
suggested = {".vmrk": partial(_read_unsupported, suggest=".vhdr"),
             ".eeg": partial(_read_unsupported, suggest=".vhdr")}

# all known file formats
readers = {**supported, **suggested}


def split_name_ext(fname):
    """Return name and supported file extension."""
    maxsuffixes = max([ext.count(".") for ext in supported])
    suffixes = Path(fname).suffixes
    for i in range(-maxsuffixes, 0):
        ext = "".join(suffixes[i:]).lower()
        if ext in readers.keys():
            return Path(fname).name[:-len(ext)], ext


def read_raw(fname, *args, **kwargs):
    """Read raw file.

    Parameters
    ----------
    fname : str
        File name to load.

    Returns
    -------
    raw : mne.io.Raw
        Raw object.

    Notes
    -----
    This function supports reading different file formats. It uses the readers dict to
    dispatch the appropriate read function for a supported file type.
    """
    _, ext = split_name_ext(fname)
    if ext is not None:
        return readers[ext](fname, *args, **kwargs)
    raise ValueError(f"Unknown file type {''.join(Path(fname).suffixes)}).")
    # here we could inspect the file signature to determine its type, which would allow us
    # to read file independently of their extensions



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QListWidget,
    QRadioButton,
)

from .utils import select_all


class PlotEvokedDialog(QDialog):
    def __init__(self, parent, channels, events, montage):
        super().__init__(parent)
        self.setWindowTitle("Plot evoked")

        grid = QGridLayout(self)
        grid.setColumnStretch(0, 2)
        grid.setColumnStretch(1, 3)

        label = QLabel("Channels:")
        label.setAlignment(Qt.AlignTop)
        grid.addWidget(label, 0, 0)
        self.picks = QListWidget()
        self.picks.insertItems(0, channels)
        self.picks.setSelectionMode(QListWidget.ExtendedSelection)
        select_all(self.picks)
        grid.addWidget(self.picks, 0, 1)

        label = QLabel("Events:")
        label.setAlignment(Qt.AlignTop)
        grid.addWidget(label, 1, 0)
        self.events = QListWidget()
        self.events.insertItems(0, events)
        self.events.setSelectionMode(QListWidget.ExtendedSelection)
        self.events.setMaximumHeight(self.events.sizeHintForRow(0) * 5.5)
        select_all(self.events)
        grid.addWidget(self.events, 1, 1)

        grid.addWidget(QLabel("Show GFP:"), 2, 0)
        self.gfp = QCheckBox()
        self.gfp.setChecked(False)
        grid.addWidget(self.gfp, 2, 1)

        self.spatial_colors_label = QLabel("Spatial colors:")
        grid.addWidget(self.spatial_colors_label, 3, 0)
        self.spatial_colors = QCheckBox()
        self.spatial_colors.setChecked(False)
        grid.addWidget(self.spatial_colors, 3, 1)

        self.topomaps = QGroupBox("Topomaps")
        self.topomaps.setCheckable(True)
        self.topomaps.setChecked(False)
        topomaps_grid = QGridLayout()
        topomaps_grid.setColumnStretch(0, 2)
        topomaps_grid.setColumnStretch(1, 3)
        self.topomaps_peaks = QRadioButton("Peaks")
        self.topomaps_auto = QRadioButton("Auto")
        self.topomaps_times = QRadioButton("Time(s):")
        self.topomaps_timelist = QLineEdit()
        topomaps_grid.addWidget(self.topomaps_peaks, 0, 0)
        topomaps_grid.addWidget(self.topomaps_auto, 1, 0)
        topomaps_grid.addWidget(self.topomaps_times, 2, 0)
        topomaps_grid.addWidget(self.topomaps_timelist, 2, 1)
        self.topomaps_peaks.setChecked(True)
        self.topomaps_timelist.setEnabled(False)
        self.topomaps.setLayout(topomaps_grid)
        grid.addWidget(self.topomaps, 4, 0, 1, 4)

        if montage is None:
            self.topomaps.setTitle("Topomaps (requires montage information)")
            self.topomaps.setCheckable(False)
            self.topomaps.setEnabled(False)
            self.topomaps.setStyleSheet("QGroupBox::title{ color: gray }")

        self.topomaps_times.toggled.connect(self.toggle_topomaps_timelist)
        self.topomaps.toggled.connect(self.toggle_ok)
        self.topomaps_times.toggled.connect(self.toggle_ok)
        self.topomaps_timelist.textChanged.connect(self.toggle_ok)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        grid.addWidget(self.buttonbox, 5, 0, 1, -1)
        self.picks.itemSelectionChanged.connect(self.toggle_ok)
        self.events.itemSelectionChanged.connect(self.toggle_ok)
        self.toggle_ok()
        grid.setSizeConstraint(QGridLayout.SetFixedSize)

    @Slot()
    def toggle_ok(self):
        if self.picks.selectedItems() and self.events.selectedItems():
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)

    @Slot()
    def toggle_topomaps_timelist(self):
        self.topomaps_timelist.setEnabled(self.topomaps_times.isChecked())


class PlotEvokedComparisonDialog(QDialog):
    def __init__(self, parent, channels, events):
        super().__init__(parent)
        self.setWindowTitle("Plot evoked comparison")

        grid = QGridLayout(self)
        grid.setColumnStretch(0, 2)
        grid.setColumnStretch(1, 3)

        label = QLabel("Channels:")
        label.setAlignment(Qt.AlignTop)
        grid.addWidget(label, 0, 0, 1, 1)
        self.picks = QListWidget()
        self.picks.insertItems(0, channels)
        self.picks.setSelectionMode(QListWidget.ExtendedSelection)
        select_all(self.picks)
        grid.addWidget(self.picks, 0, 1, 1, 1)

        label = QLabel("Events:")
        label.setAlignment(Qt.AlignTop)
        grid.addWidget(label, 1, 0, 1, 1)
        self.events = QListWidget()
        self.events.insertItems(0, events)
        self.events.setSelectionMode(QListWidget.ExtendedSelection)
        self.events.setMaximumHeight(self.events.sizeHintForRow(0) * 5.5)
        select_all(self.events)
        grid.addWidget(self.events, 1, 1, 1, 1)

        grid.addWidget(QLabel("Average epochs:"), 2, 0)
        self.average_epochs = QComboBox()
        self.average_epochs.addItems(["mean", "median"])
        self.average_epochs.setCurrentIndex(0)
        grid.addWidget(self.average_epochs, 2, 1)

        grid.addWidget(QLabel("Combine channels:"), 3, 0)
        self.combine_channels = QComboBox()
        self.combine_channels.addItems(["gfp", "mean", "median", "std"])
        self.combine_channels.setCurrentIndex(0)
        grid.addWidget(self.combine_channels, 3, 1)

        grid.addWidget(QLabel("Confidence intervals:"), 4, 0)
        self.confidence_intervals = QCheckBox()
        self.confidence_intervals.setChecked(True)
        grid.addWidget(self.confidence_intervals, 4, 1)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        grid.addWidget(self.buttonbox, 5, 0, 1, -1)
        self.picks.itemSelectionChanged.connect(self.toggle_ok)
        self.events.itemSelectionChanged.connect(self.toggle_ok)
        self.toggle_ok()
        grid.setSizeConstraint(QGridLayout.SetFixedSize)

    @Slot()
    def toggle_ok(self):
        if self.picks.selectedItems() and self.events.selectedItems():
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)


class PlotEvokedTopomaps(QDialog):
    def __init__(self, parent, events):
        super().__init__(parent)
        self.setWindowTitle("Plot evoked topomaps")

        grid = QGridLayout(self)
        grid.setColumnStretch(0, 2)
        grid.setColumnStretch(1, 3)

        label = QLabel("Events:")
        label.setAlignment(Qt.AlignTop)
        grid.addWidget(label, 0, 0)
        self.events = QListWidget()
        self.events.insertItems(0, events)
        self.events.setSelectionMode(QListWidget.ExtendedSelection)
        self.events.setMaximumHeight(self.events.sizeHintForRow(0) * 5.5)
        select_all(self.events)
        grid.addWidget(self.events, 0, 1)

        grid.addWidget(QLabel("Average epochs:"), 1, 0)
        self.average_epochs = QComboBox()
        self.average_epochs.addItems(["mean", "median"])
        self.average_epochs.setCurrentIndex(0)
        grid.addWidget(self.average_epochs, 1, 1)

        timepoints = QGroupBox("Select time point(s):")
        timepoints_grid = QGridLayout()
        timepoints_grid.setColumnStretch(0, 2)
        timepoints_grid.setColumnStretch(1, 3)
        timepoints.setLayout(timepoints_grid)
        self.auto = QRadioButton("Auto")
        self.peaks = QRadioButton("Peaks")
        self.interactive = QRadioButton("Interactive")
        self.manual = QRadioButton("Manual:")
        self.timelist = QLineEdit()
        self.auto.setChecked(True)
        timepoints_grid.addWidget(self.auto, 0, 0)
        timepoints_grid.addWidget(self.peaks, 1, 0)
        timepoints_grid.addWidget(self.interactive, 2, 0)
        timepoints_grid.addWidget(self.manual, 3, 0)
        timepoints_grid.addWidget(self.timelist, 3, 1)
        self.timelist.setEnabled(False)
        self.manual.toggled.connect(self.toggle_timelist)
        grid.addWidget(timepoints, 2, 0, 1, 2)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        grid.addWidget(self.buttonbox, 3, 0, 1, -1)
        self.events.itemSelectionChanged.connect(self.toggle_ok)
        self.toggle_ok()
        grid.setSizeConstraint(QGridLayout.SetFixedSize)

    @Slot()
    def toggle_ok(self):
        enable = bool(self.events.selectedItems())
        self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(enable)

    @Slot()
    def toggle_timelist(self):
        self.timelist.setEnabled(self.manual.isChecked())



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QTableWidgetItem


def select_all(list_widget):
    """Select all items in a QListWidget."""
    for i in range(list_widget.count()):
        list_widget.item(i).setSelected(True)


class IntTableWidgetItem(QTableWidgetItem):
    def __init__(self, value):
        super().__init__(str(value))

    def __lt__(self, other):
        return int(self.data(Qt.EditRole)) < int(other.data(Qt.EditRole))

    def setData(self, role, value):
        try:
            value = int(value)
        except ValueError:
            return
        else:
            if value >= 0:  # event position and type must not be negative
                super().setData(role, str(value))

    def value(self):
        return int(self.data(Qt.DisplayRole))


class FloatTableWidgetItem(QTableWidgetItem):
    def __init__(self, value):
        super().__init__(str(value))

    def __lt__(self, other):
        return float(self.data(Qt.EditRole)) < float(other.data(Qt.EditRole))

    def setData(self, role, value):
        try:
            value = float(value)
        except ValueError:
            return
        else:
            if value >= 0:  # event position and type must not be negative
                super().setData(role, str(value))

    def value(self):
        return float(self.data(Qt.DisplayRole))



# © MNELAB developers
#
# License: BSD (3-clause)

import math

import matplotlib.pyplot as plt
import numpy as np
from mne.stats import permutation_cluster_1samp_test as pcluster_test
from mne.time_frequency import tfr_multitaper
from mne.viz import plot_compare_evokeds


def _center_cmap(cmap, vmin, vmax, name="cmap_centered"):
    """
    Center given colormap (ranging from vmin to vmax) at value 0.

    Taken from MNE-Python v0.24, as it will be removed in MNE-Python v1.0.

    Parameters
    ----------
    cmap : matplotlib.colors.Colormap
        The colormap to center around 0.
    vmin : float
        Minimum value in the data to map to the lower end of the colormap.
    vmax : float
        Maximum value in the data to map to the upper end of the colormap.
    name : str
        Name of the new colormap. Defaults to 'cmap_centered'.

    Returns
    -------
    cmap_centered : matplotlib.colors.Colormap
        The new colormap centered around 0.

    Notes
    -----
    This function can be used in situations where vmin and vmax are not symmetric around
    zero. Normally, this results in the value zero not being mapped to white anymore in many
    colormaps. Using this function, the value zero will be mapped to white even for
    asymmetric positive and negative value ranges. Note that this could also be achieved by
    re-normalizing a given colormap by subclassing matplotlib.colors.Normalize as described
    here:
    https://matplotlib.org/users/colormapnorms.html#custom-normalization-two-linear-ranges
    """
    from matplotlib.colors import LinearSegmentedColormap

    vzero = abs(vmin) / float(vmax - vmin)
    index_old = np.linspace(0, 1, cmap.N)
    index_new = np.hstack([
        np.linspace(0, vzero, cmap.N // 2, endpoint=False),
        np.linspace(vzero, 1, cmap.N // 2),
    ])

    colors = ("red", "green", "blue", "alpha")
    cdict = {name: [] for name in colors}
    for old, new in zip(index_old, index_new):
        for color, name in zip(cmap(old), colors):
            cdict[name].append((new, color, color))
    return LinearSegmentedColormap(name, cdict)


def _get_rows_cols(n):
    if n <= 3:
        rows, cols = 1, n
    else:
        rows = round(math.sqrt(n))
        cols = math.ceil(math.sqrt(n))
    return rows, cols


def _calc_tfr(epochs, freqs, baseline, times, alpha=None):
    """
    Calculate AverageTFR and significance masks for given epochs.

    Adapted from https://mne.tools/dev/auto_examples/time_frequency/time_frequency_erds.html

    Parameters
    ----------
    epochs : mne.epochs.Epochs
        Epochs extracted from a Raw instance.
    freqs : np.ndarray
        The frequencies in Hz.
    baseline : array_like, shape (2,)
        The time interval to apply rescaling / baseline correction.
    times : array_like, shape (2,)
        Start and end of crop time interval.
    alpha : float, optional
        If specified, calculate significance maps with threshold `alpha`, by default `None`.

    Returns
    -------
    dict[str, tuple[mne.time_frequency.tfr.EpochsTFR, dict[str, np.ndarray | None]]]
        A dictionary where keys are event IDs and values are tuples (`tfr_ev`, `masks`).
        `tfr_ev` is the EpochsTFR object for the respective event. `masks` is again a
        dictionary, where keys are channel names and values are significance masks.
        Significance masks are `None` if `alpha` was not specified.
    """
    tfr = tfr_multitaper(epochs, freqs, freqs, average=False, return_itc=False)
    tfr.apply_baseline(baseline, mode="percent")
    tfr.crop(*times)

    pcluster_kwargs = dict(
        n_permutations=100,
        step_down_p=0.05,
        seed=1,
        buffer_size=None,
        out_type='mask',
    )

    res = {}

    for event in epochs.event_id:
        tfr_ev = tfr[event]
        masks = {}
        for ch in range(epochs.info["nchan"]):
            mask = None
            if alpha is not None:
                # positive clusters
                _, c1, p1, _ = pcluster_test(tfr_ev.data[:, ch], tail=1, **pcluster_kwargs)
                # negative clusters
                _, c2, p2, _ = pcluster_test(tfr_ev.data[:, ch], tail=-1, **pcluster_kwargs)

                c = np.stack(c1 + c2, axis=2)  # combined clusters
                p = np.concatenate((p1, p2))   # combined p-values
                mask = c[..., p <= alpha].any(axis=-1)
            masks[epochs.ch_names[ch]] = mask
        res[event] = (tfr_ev, masks)
    return res


def plot_erds(tfr_and_masks):
    """
    Plot ERDS maps from given TFR and significance masks.

    Parameters
    ----------
    tfr_and_masks : dict[str, tuple[EpochsTFR, dict[str, np.ndarray |None]]]
        A dictionary where keys are event IDs and values are tuples (`tfr_ev`, `masks`).
        `tfr_ev` is the EpochsTFR object for the respective event. `masks` is again a
        dictionary, where keys are channel names and values are significance masks.

    Returns
    -------
    list[matplotlib.figure.Figure]
        A list of the figure(s) generated, one figure per event.
    """
    figs = []

    for event, (tfr_ev, masks) in tfr_and_masks.items():
        n_rows, n_cols = _get_rows_cols(tfr_ev.info["nchan"])
        widths = n_cols * [10] + [1]  # each map has width 10, each colorbar width 1
        fig, axes = plt.subplots(n_rows, n_cols + 1, gridspec_kw={"width_ratios": widths})
        vmin, vmax = -1, 2  # default for ERDS maps
        cmap = _center_cmap(plt.cm.RdBu, vmin, vmax)

        # skip the last column in `axes`, as it contains the colorbar
        for (ch_name, mask), ax in zip(masks.items(), axes[..., :-1].flat):
            tfr_ev.average().plot(
                [ch_name],
                vmin=vmin,
                vmax=vmax,
                cmap=(cmap, False),
                axes=ax,
                colorbar=False,
                mask=mask,
                mask_style="mask" if mask is not None else None,  # avoid RuntimeWarning
                show=False,
            )
            ax.set_title(ch_name, fontsize=10)
            ax.axvline(0, linewidth=1, color="black", linestyle=":")
            ax.set(xlabel="t (s)", ylabel="f (Hz)")
            ax.label_outer()
        for ax in axes[..., -1].flat:
            fig.colorbar(axes.flat[0].images[-1], cax=ax)
        fig.suptitle(f"ERDS – {event}")
        figs.append(fig)
    return figs


def plot_erds_topomaps(epochs, events, freqs, baseline, times):
    """
    Plot ERDS topomaps, one figure per event.

    Parameters
    ----------
    epochs : mne.epochs.Epochs
        Epochs extracted from a Raw instance.
    events : list[str]
        Events to include.
    freqs : np.ndarray
        Array of frequencies over which the average is taken.
    baseline : tuple[float, float]
        Start and end times for baseline correction.
    times : tuple[float, float]
        Start and end times between which the average is taken.

    Returns
    -------
    list[matplotlib.figure.Figure]
        A list of the figure(s) generated.
    """
    vmin, vmax = -1, 2
    cmap = _center_cmap(plt.cm.RdBu, vmin, vmax)

    figs = []
    for event in events:
        tfr = tfr_multitaper(epochs[event], freqs, freqs, average=True, return_itc=False)
        tfr.apply_baseline(baseline, mode="percent")
        tfr.crop(*times)
        fig = tfr.plot_topomap(
            title=f"Event: {event}",
            unit="ERDS",
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            cbar_fmt="%.1f",
        )
        fig.set_size_inches(4, 3)
        fig.set_tight_layout(True)
        figs.append(fig)
    return figs


def plot_evoked(
    epochs,
    picks,
    events,
    gfp,
    spatial_colors,
    topomap_times,
):
    """
    Plot evoked potentials of different events for individual channels.

    If multiple events are selected, one figure will be returned for each.

    Parameters
    ----------
    epochs : mne.epochs.Epochs
        Epochs extracted from a Raw instance.
    picks : list[str]
        Channels to include.
    events : list[str]
        Events to include.
    gfp : bool | "only"
        Plot the global field power (GFP).
    spatial_colors : bool
        If `True`, the lines are color coded by mapping physical sensor
        coordinates into color values. Spatially similar channels will have
        similar colors. Bad channels will be dotted. If `False`, the good
        channels are plotted black and bad channels red.
    topomap_times : list[float] | "auto" | "peaks"
        The time point(s) to plot. If `"auto"`, 5 evenly spaced topographies
        between the first and last time instant will be shown. If `"peaks"`,
        finds time points automatically by checking for 3 local maxima in
        Global Field Power.

    Returns
    -------
    list[matplotlib.figure.Figure]
        A list of the figure(s) generated.
    """
    figs = []
    for event in events:
        evoked = epochs[event].average(picks=picks)
        if topomap_times:
            figs.append(evoked.plot_joint(
                times=topomap_times,
                title=f'Event: {event}',
                picks=picks,
                ts_args={
                    "spatial_colors": spatial_colors,
                    "gfp": gfp,
                }
            ))
        else:
            figs.append(evoked.plot(
                window_title=f'Event: {event}',
                picks=picks,
                spatial_colors=spatial_colors,
                gfp=gfp
            ))
    return figs


def plot_evoked_comparison(
    epochs,
    picks,
    events,
    average_method,
    combine,
    confidence_intervals,
):
    """
    Plot evoked potentials of different events averaged over channels.

    If multiple channel types are selected, one figure will be returned for
    each channel type.

    Parameters
    ----------
    epochs : mne.epochs.Epochs
        Epochs extracted from a Raw instance.
    picks : list[str]
        Channels to include.
    events : list[str]
        Events to include.
    average_method : {"mean", "median"}
        How to combine the data during averaging.
    combine : {"gfp", "std", mean", "median"}
        How to combine information across channels.
    confidence_intervals : bool
        If `True`, plot confidence intervals as shaded areas.

    Returns
    -------
    list[matplotlib.figure.Figure]
        A list of the figure(s) generated.
    """
    if confidence_intervals:
        evokeds = {e: list(epochs[e].iter_evoked()) for e in events}
    else:
        evokeds = {e: epochs[e].average(picks=picks, method=average_method, by_event_type=True) for e in events}  # noqa: E501
    return plot_compare_evokeds(evokeds, picks=picks, combine=combine)


def plot_evoked_topomaps(epochs, events, average_method, times):
    """
    Plot evoked topomaps.

    One figure is generated for each event.

    Parameters
    ----------
    epochs : mne.epochs.Epochs
        Epochs extracted from a Raw instance.
    events : list[str]
        Events to include.
    average_method : "mean" | "median
        How to average epochs.
    times : list[float] | "auto" | "peaks" | "interactive"
        The time point(s) to plot.

    Returns
    -------
    list[matplotlib.figure.Figure]
        A list of the figure(s) generated.
    """
    figs = []
    for event in events:
        evoked = epochs[event].average(method=average_method)
        figs.append(evoked.plot_topomap(times, title=f'Event: {event}'))
        if times == 'interactive':
            figs[-1].set_size_inches(6, 4)
    return figs



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
)


class InterpolateBadsDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Interpolate bad channels")
        vbox = QVBoxLayout(self)
        grid = QGridLayout()
        grid.addWidget(QLabel("Reset bads:"), 0, 0)
        self.reset_bads_checkbox = QCheckBox()
        self.reset_bads_checkbox.setChecked(True)
        grid.addWidget(self.reset_bads_checkbox, 0, 1)
        grid.addWidget(QLabel("Mode:"), 1, 0)
        self.mode_select = QComboBox()
        self.modes = {"Accurate": "accurate", "Fast": "fast"}
        self.mode_select.addItems(self.modes.keys())
        self.mode_select.setCurrentText("Accurate")
        grid.addWidget(self.mode_select, 1, 1)
        grid.addWidget(QLabel("Origin (x, y, z):"), 2, 0)
        hbox = QHBoxLayout()
        self.x = QDoubleSpinBox()
        self.x.setValue(0)
        self.x.setDecimals(3)
        hbox.addWidget(self.x)
        self.y = QDoubleSpinBox()
        self.y.setValue(0)
        self.y.setDecimals(3)
        hbox.addWidget(self.y)
        self.z = QDoubleSpinBox()
        self.z.setValue(0.04)
        self.z.setDecimals(3)
        hbox.addWidget(self.z)
        grid.addLayout(hbox, 2, 1)

        vbox.addLayout(grid)
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(buttonbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    @property
    def origin(self):
        x = float(self.x.value())
        y = float(self.y.value())
        z = float(self.z.value())
        return x, y, z

    @property
    def mode(self):
        return self.mode_select.currentText()

    @property
    def reset_bads(self):
        return self.reset_bads_checkbox.isChecked()



# © MNELAB developers
#
# License: BSD (3-clause)

from mne.channels.channels import _unit2human
from mne.io.pick import get_channel_type_constants
from PySide6.QtCore import Slot
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QVBoxLayout,
)


def _label(channel_type):
    label = channel_type.upper()
    try:
        unit = _unit2human[get_channel_type_constants()[channel_type]["unit"]]
        label += f" ({unit})"
    except KeyError:
        pass
    return label


class DropBadEpochsDialog(QDialog):
    def __init__(self, parent, types):
        super().__init__(parent)
        self.setWindowTitle("Drop bad epochs")

        vbox = QVBoxLayout(self)

        self.reject_box = QGroupBox("Reject (maximum PTP amplitude)")
        self.reject_box.setCheckable(True)
        self.reject_box.setChecked(False)
        self.reject_fields = {}
        reject_grid = QGridLayout()
        for row, type in enumerate(types):
            reject_grid.addWidget(QLabel(_label(type)), row, 0)
            self.reject_fields[type] = QLineEdit()
            reject_grid.addWidget(self.reject_fields[type], row, 1)
        self.reject_box.setLayout(reject_grid)
        vbox.addWidget(self.reject_box)

        self.flat_box = QGroupBox("Flat (minimum PTP amplitude)")
        self.flat_box.setCheckable(True)
        self.flat_box.setChecked(False)
        self.flat_fields = {}
        flat_grid = QGridLayout()
        for row, type in enumerate(types):
            flat_grid.addWidget(QLabel(_label(type)), row, 0)
            self.flat_fields[type] = QLineEdit()
            flat_grid.addWidget(self.flat_fields[type], row, 1)
        self.flat_box.setLayout(flat_grid)
        vbox.addWidget(self.flat_box)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        self.flat_box.toggled.connect(self.toggle_ok)
        self.reject_box.toggled.connect(self.toggle_ok)
        vbox.setSizeConstraint(QGridLayout.SetFixedSize)

    @Slot()
    def toggle_ok(self):
        self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
        if self.reject_box.isChecked() or self.flat_box.isChecked():
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)



# © MNELAB developers
#
# License: BSD (3-clause)

import multiprocessing as mp
import sys
import traceback
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from sys import version_info

import mne
import numpy as np
from mne.io.pick import channel_type
from PySide6.QtCore import QEvent, QMetaObject, QModelIndex, QObject, Qt, QUrl, Slot
from PySide6.QtGui import QAction, QDesktopServices, QDropEvent, QIcon, QKeySequence
from PySide6.QtWidgets import (
    QApplication,
    QFileDialog,
    QFrame,
    QLabel,
    QListWidget,
    QMainWindow,
    QMessageBox,
    QSplitter,
)
from pyxdf import resolve_streams

from .dialogs import *  # noqa: F403
from .io import writers
from .io.mat import parse_mat
from .io.xdf import get_xml, list_chunks
from .model import InvalidAnnotationsError, LabelsNotFoundError, Model
from .settings import SettingsDialog, read_settings, write_settings
from .utils import count_locations, have, image_path, interface_style, natural_sort
from .viz import (
    _calc_tfr,
    plot_erds,
    plot_erds_topomaps,
    plot_evoked,
    plot_evoked_comparison,
    plot_evoked_topomaps,
)
from .widgets import InfoWidget


class MainWindow(QMainWindow):
    """MNELAB main window."""
    def __init__(self, model: Model):
        """Initialize MNELAB main window.

        Parameters
        ----------
        model : mnelab.model.Model instance
            The main window needs to connect to a model containing all data sets. This
            decouples the GUI from the data (model/view).
        """
        super().__init__()
        self.model = model  # data model
        self.setWindowTitle("MNELAB")
        sys.excepthook = self._excepthook

        # restore settings
        settings = read_settings()
        self.recent = settings["recent"]  # list of recent files
        self.resize(settings["size"])
        self.move(settings["pos"])

        # remove None entries from self.recent
        self.recent = [recent for recent in self.recent if recent is not None]

        # trigger theme setting
        QIcon.setThemeSearchPaths([str(Path(__file__).parent / "icons")])
        self.event(QEvent(QEvent.PaletteChange))

        self.actions = {}  # contains all actions

        # initialize menus
        file_menu = self.menuBar().addMenu("&File")
        icon = QIcon.fromTheme("open-file")
        self.actions["open_file"] = file_menu.addAction(icon, "&Open...", self.open_data,
                                                        QKeySequence.Open)
        self.recent_menu = file_menu.addMenu("Open recent")
        self.recent_menu.aboutToShow.connect(self._update_recent_menu)
        self.recent_menu.triggered.connect(self._load_recent)
        if not self.recent:
            self.recent_menu.setEnabled(False)
        self.actions["close_file"] = file_menu.addAction("&Close", self.model.remove_data,
                                                         QKeySequence.Close)
        self.actions["close_all"] = file_menu.addAction("Close all", self.close_all)
        file_menu.addSeparator()
        self.actions["import_bads"] = file_menu.addAction(
            "Import bad channels...",
            lambda: self.import_file(model.import_bads, "Import bad channels", "*.csv")
        )
        self.actions["import_events"] = file_menu.addAction(
            "Import events...",
            lambda: self.import_file(model.import_events, "Import events", "*.csv *.fif"),
        )
        self.actions["import_annotations"] = file_menu.addAction(
            "Import annotations...",
            lambda: self.import_file(model.import_annotations, "Import annotations",
                                     "*.csv")
        )
        self.actions["import_ica"] = file_menu.addAction(
            "Import &ICA...",
            lambda: self.open_file(model.import_ica, "Import ICA", "*.fif *.fif.gz")
        )
        file_menu.addSeparator()
        self.export_menu = file_menu.addMenu("Export data")
        for ext, description in writers.items():
            action = "export_data" + ext.replace(".", "_")
            self.actions[action] = self.export_menu.addAction(
                f"{ext[1:].upper()} ({description[1]})...",
                partial(self.export_file, model.export_data, "Export data", "*" + ext)
            )
        self.actions["export_bads"] = file_menu.addAction(
            "Export &bad channels...",
            lambda: self.export_file(model.export_bads, "Export bad channels", "*.csv")
        )
        self.actions["export_events"] = file_menu.addAction(
            "Export &events...",
            lambda: self.export_file(model.export_events, "Export events", "*.csv")
        )
        self.actions["export_annotations"] = file_menu.addAction(
            "Export &annotations...",
            lambda: self.export_file(model.export_annotations, "Export annotations",
                                     "*.csv")
        )
        self.actions["export_ica"] = file_menu.addAction(
            "Export ICA...",
            lambda: self.export_file(model.export_ica, "Export ICA", "*.fif *.fif.gz")
        )
        file_menu.addSeparator()
        icon = QIcon.fromTheme("xdf-metadata")
        self.actions["xdf_metadata"] = file_menu.addAction(
            "Show XDF metadata",
            self.xdf_metadata,
        )
        self.actions["xdf_chunks"] = file_menu.addAction("Inspect XDF chunks...",
                                                         self.xdf_chunks)
        file_menu.addSeparator()
        self.actions["settings"] = file_menu.addAction(
            "Settings...",
            SettingsDialog(self).exec,
        )
        file_menu.addSeparator()
        self.actions["quit"] = file_menu.addAction("&Quit", self.close, QKeySequence.Quit)

        edit_menu = self.menuBar().addMenu("&Edit")
        self.actions["pick_chans"] = edit_menu.addAction("P&ick channels...",
                                                         self.pick_channels)
        self.actions["rename_channels"] = edit_menu.addAction(
            "Rename channels...",
            self.rename_channels,
        )
        self.actions["chan_props"] = edit_menu.addAction(
            QIcon.fromTheme("chan-props"),
            "Edit channel &properties...",
            self.channel_properties,
        )
        edit_menu.addSeparator()
        self.actions["set_montage"] = edit_menu.addAction("Set &montage...",
                                                          self.set_montage)
        self.actions["clear_montage"] = edit_menu.addAction(
            "Clear montage",
            self.clear_montage,
        )
        edit_menu.addSeparator()
        self.actions["change_ref"] = edit_menu.addAction(
            "Change &reference...",
            self.change_reference,
        )
        edit_menu.addSeparator()
        self.actions["annotations"] = edit_menu.addAction(
            "Edit &Annotations...",
            self.edit_annotations,
        )
        self.actions["events"] = edit_menu.addAction(
            "Edit &Events...",
            self.edit_events,
        )

        edit_menu.addSeparator()
        self.actions["crop"] = edit_menu.addAction("&Crop data...", self.crop)
        self.actions["append_data"] = edit_menu.addAction("Appen&d data...",
                                                          self.append_data)

        plot_menu = self.menuBar().addMenu("&Plot")
        self.actions["plot_data"] = plot_menu.addAction(
            QIcon.fromTheme("plot-data"),
            "Plot &Data",
            self.plot_data,
        )
        self.actions["plot_psd"] = plot_menu.addAction(
            QIcon.fromTheme("plot-psd"),
            "Plot &PSD",
            self.plot_psd,
        )
        plot_menu.addSeparator()
        self.actions["plot_locations"] = plot_menu.addAction(
            QIcon.fromTheme("plot-locations"),
            "Plot &channel locations",
            self.plot_locations,
        )
        plot_menu.addSeparator()
        self.actions["plot_erds"] = plot_menu.addAction(
            "Plot &ERDS maps...",
            self.plot_erds,
        )
        self.actions["plot_erds_topomaps"] = plot_menu.addAction(
            "Plot ERDS topomaps...",
            self.plot_erds_topomaps,
        )
        plot_menu.addSeparator()
        self.actions["plot_evoked"] = plot_menu.addAction(
            "Plot evoked...",
            self.plot_evoked,
        )
        self.actions["plot_evoked_comparison"] = plot_menu.addAction(
            "Plot evoked comparison...",
            self.plot_evoked_comparison,
        )
        self.actions["plot_evoked_topomaps"] = plot_menu.addAction(
            "Plot evoked topomaps...",
            self.plot_evoked_topomaps,
        )
        plot_menu.addSeparator()
        self.actions["plot_ica_components"] = plot_menu.addAction(
            "Plot ICA &components",
            self.plot_ica_components,
        )
        self.actions["plot_ica_sources"] = plot_menu.addAction(
            "Plot ICA &sources",
            self.plot_ica_sources,
        )

        tools_menu = self.menuBar().addMenu("&Tools")
        icon = QIcon.fromTheme("filter-data")
        self.actions["filter"] = tools_menu.addAction(icon, "&Filter data...",
                                                      self.filter_data)
        icon = QIcon.fromTheme("find-events")
        self.actions["find_events"] = tools_menu.addAction(icon, "Find &events...",
                                                           self.find_events)
        self.actions["events_from_annotations"] = tools_menu.addAction(
            "Create events from annotations", self.events_from_annotations
        )
        self.actions["annotations_from_events"] = tools_menu.addAction(
            "Create annotations from events", self.annotations_from_events
        )
        tools_menu.addSeparator()
        self.actions["convert_od"] = tools_menu.addAction(
            "Convert to &optical density",
            self.convert_od,
        )
        self.actions["convert_bl"] = tools_menu.addAction(
            "Convert to &haemoglobin",
            self.convert_bl,
        )
        tools_menu.addSeparator()

        icon = QIcon.fromTheme("run-ica")
        self.actions["run_ica"] = tools_menu.addAction(icon, "Run &ICA...", self.run_ica)
        self.actions["apply_ica"] = tools_menu.addAction("Apply &ICA", self.apply_ica)
        tools_menu.addSeparator()
        self.actions["interpolate_bads"] = tools_menu.addAction(
            "Interpolate bad channels...", self.interpolate_bads
        )
        tools_menu.addSeparator()
        icon = QIcon.fromTheme("epoch-data")
        self.actions["epoch_data"] = tools_menu.addAction(icon, "Create epochs...",
                                                          self.epoch_data)
        self.actions["drop_bad_epochs"] = tools_menu.addAction(
            "Drop bad epochs...",
            self.drop_bad_epochs,
        )

        view_menu = self.menuBar().addMenu("&View")
        self.actions["history"] = view_menu.addAction("&History", self.show_history)
        self.actions["toolbar"] = view_menu.addAction("&Toolbar", self._toggle_toolbar)
        self.actions["toolbar"].setCheckable(True)
        self.actions["statusbar"] = view_menu.addAction("&Statusbar",
                                                        self._toggle_statusbar)
        self.actions["statusbar"].setCheckable(True)

        help_menu = self.menuBar().addMenu("&Help")
        self.actions["about"] = help_menu.addAction("&About", self.show_about)
        self.actions["about_qt"] = help_menu.addAction("About &Qt", self.show_about_qt)
        help_menu.addSeparator()
        self.actions["documentation"] = help_menu.addAction(
            "&Documentation",
            self.show_documentation,
        )
        # actions that are always enabled
        self.always_enabled = ["open_file", "about", "about_qt", "quit", "xdf_chunks",
                               "toolbar", "statusbar", "settings", "documentation",
                               "history"]

        # set up toolbar
        self.toolbar = self.addToolBar("toolbar")
        self.toolbar.setObjectName("toolbar")
        self.toolbar.addAction(self.actions["open_file"])
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.actions["chan_props"])
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.actions["plot_data"])
        self.toolbar.addAction(self.actions["plot_psd"])
        self.toolbar.addAction(self.actions["plot_locations"])
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.actions["filter"])
        self.toolbar.addAction(self.actions["find_events"])
        self.toolbar.addAction(self.actions["epoch_data"])
        self.toolbar.addAction(self.actions["run_ica"])
        self.toolbar.setMovable(False)
        self.setUnifiedTitleAndToolBarOnMac(True)
        if settings["toolbar"]:
            self.toolbar.show()
            self.actions["toolbar"].setChecked(True)
        else:
            self.toolbar.hide()
            self.actions["toolbar"].setChecked(False)

        # set up data model for sidebar (list of open files)
        self.sidebar = QListWidget()
        self.sidebar.setFrameStyle(QFrame.NoFrame)
        self.sidebar.setFocusPolicy(Qt.NoFocus)
        self.sidebar.setAcceptDrops(True)
        self.sidebar.setDragEnabled(True)
        self.sidebar.setDragDropMode(QListWidget.InternalMove)
        self.sidebar.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.sidebar.setEditTriggers(QListWidget.DoubleClicked)
        self.sidebar.model().rowsMoved.connect(self._sidebar_move_event)
        self.sidebar.itemDelegate().commitData.connect(self._sidebar_edit_event)
        self.sidebar.clicked.connect(self._update_data)

        splitter = QSplitter()
        splitter.addWidget(self.sidebar)

        self.infowidget = InfoWidget()
        splitter.addWidget(self.infowidget)
        width = splitter.size().width()
        splitter.setSizes((int(width * 0.3), int(width * 0.7)))
        self.setCentralWidget(splitter)

        self.status_label = QLabel()
        self.statusBar().addPermanentWidget(self.status_label)
        if settings["statusbar"]:
            self.statusBar().show()
            self.actions["statusbar"].setChecked(True)
        else:
            self.statusBar().hide()
            self.actions["statusbar"].setChecked(False)

        self.setAcceptDrops(True)
        self.data_changed()

    def _excepthook(self, type, value, traceback_):
        exception_text = str(value)
        traceback_text = "".join(traceback.format_exception(type, value, traceback_))
        print(traceback_text, file=sys.stderr)
        ErrorMessageBox(self, exception_text, "", traceback_text).show()

    def _sidebar_edit_event(self, edit):
        """
        Triggered when a data set in the sidebar is renamed.

        Parameters
        ----------
        edit : PySide6.QtWidgets.QLineEdit
            The text editor.
        """
        self.model.current["name"] = edit.text()

    def _sidebar_move_event(self, parent, start, end, destination, row):
        """
        Triggered when an item in the sidebar is moved.

        Parameters
        ----------
        parent : PySide6.QtCore.QModelIndex
            Unused.
        start : int
            The source index of the item.
        end : int
            Unused (equals start as the sidebar only allows single selection).
        destination : PySide6.QtCore.QModelIndex
            Unused.
        row : int
            The target index.
        """
        self.model.move_data(start, row)

    @contextmanager
    def _wait_cursor(self):
        # disabled on macOS because of outdated icon
        if sys.platform.startswith("darwin"):
            yield
        else:
            default_cursor = self.cursor()
            self.setCursor(Qt.WaitCursor)
            try:
                yield
            finally:
                self.setCursor(default_cursor)

    def data_changed(self):
        # update sidebar
        self.sidebar.clear()
        self.sidebar.insertItems(0, self.model.names)
        self.sidebar.setCurrentRow(self.model.index)
        for it in range(self.sidebar.count()):
            item = self.sidebar.item(it)
            item.setFlags(item.flags() | Qt.ItemIsEditable)

        # update info widget
        if self.model.data:
            self.infowidget.set_values(self.model.get_info())
        else:
            self.infowidget.clear()

        # update status bar
        if self.model.data:
            mb = self.model.nbytes / 1024 ** 2
            self.status_label.setText(f"Total Memory: {mb:.2f} MB")
        else:
            self.status_label.clear()

        # toggle actions
        if len(self.model) == 0:  # disable if no data sets are currently open
            enabled = False
        else:
            enabled = True

        for name, action in self.actions.items():  # toggle
            if name not in self.always_enabled:
                action.setEnabled(enabled)

        if self.model.data:  # toggle if specific conditions are met
            bads = bool(self.model.current["data"].info["bads"])
            self.actions["export_bads"].setEnabled(enabled and bads)
            events = len(self.model.current["events"]) > 0
            self.actions["export_events"].setEnabled(enabled and events)
            if self.model.current["dtype"] == "raw":
                annot = bool(self.model.current["data"].annotations)
            else:
                annot = False
            self.actions["export_annotations"].setEnabled(enabled and annot)
            self.actions["annotations"].setEnabled(enabled)
            locations = count_locations(self.model.current["data"].info)
            self.actions["plot_locations"].setEnabled(enabled and locations)
            ica = bool(self.model.current["ica"])
            self.actions["apply_ica"].setEnabled(enabled and ica)
            self.actions["export_ica"].setEnabled(enabled and ica)
            self.actions["plot_erds"].setEnabled(
                enabled and self.model.current["dtype"] == "epochs"
            )
            self.actions["plot_erds_topomaps"].setEnabled(
                enabled and locations and self.model.current["dtype"] == "epochs"
            )
            self.actions["plot_evoked"].setEnabled(
                enabled and self.model.current["dtype"] == "epochs"
            )
            self.actions["plot_evoked_comparison"].setEnabled(
                enabled and self.model.current["dtype"] == "epochs"
            )
            self.actions["plot_evoked_topomaps"].setEnabled(
                enabled and locations and self.model.current["dtype"] == "epochs"
            )
            self.actions["plot_ica_components"].setEnabled(enabled and ica and locations)
            self.actions["plot_ica_sources"].setEnabled(enabled and ica)
            self.actions["interpolate_bads"].setEnabled(enabled and locations and bads)
            self.actions["events"].setEnabled(enabled)
            self.actions["events_from_annotations"].setEnabled(enabled and annot)
            self.actions["annotations_from_events"].setEnabled(enabled and events)
            self.actions["find_events"].setEnabled(
                enabled and self.model.current["dtype"] == "raw"
            )
            self.actions["epoch_data"].setEnabled(
                enabled and events and self.model.current["dtype"] == "raw"
            )
            self.actions["drop_bad_epochs"].setEnabled(
                enabled and events and self.model.current["dtype"] == "epochs"
            )
            self.actions["clear_montage"].setEnabled(
                enabled and self.model.current["montage"] is not None
            )
            self.actions["crop"].setEnabled(
                enabled and self.model.current["dtype"] == "raw"
            )
            append = bool(self.model.get_compatibles())
            self.actions["append_data"].setEnabled(
                enabled and append and (self.model.current["dtype"] in ("raw", "epochs"))
            )
            self.actions["xdf_metadata"].setEnabled(
                enabled and self.model.current["ftype"] in ["XDF", "XDFZ", "XDF.GZ"]
            )
            self.actions["convert_od"].setEnabled(
                len(
                    mne.pick_types(
                        self.model.current["data"].info, fnirs="fnirs_cw_amplitude"
                    )
                )
            )
            self.actions["convert_bl"].setEnabled(
                len(mne.pick_types(self.model.current["data"].info, fnirs="fnirs_od"))
            )
            # disable unsupported exporters for epochs (all must support raw)
            if self.model.current["dtype"] == "epochs":
                for ext, description in writers.items():
                    action = "export_data" + ext.replace(".", "_")
                    if "epoch" in description[2]:
                        self.actions[action].setEnabled(True)
                    else:
                        self.actions[action].setEnabled(False)
        # add to recent files
        if len(self.model) > 0:
            self._add_recent(self.model.current["fname"])

    def open_data(self, fname=None):
        """Open raw file."""
        if fname is None:
            # getOpenFileNames returns a tuple (filenames, selected_filter)
            fnames, _ = QFileDialog.getOpenFileNames(self, "Open raw")
        else:
            fnames = [fname]
        for fname in fnames:
            if not (Path(fname).is_file() or Path(fname).is_dir()):
                self._remove_recent(fname)
                QMessageBox.critical(self, "File does not exist",
                                     f"File {fname} does not exist anymore.")
                return

            ext = "".join(Path(fname).suffixes)

            if any([ext.endswith(e) for e in (".xdf", ".xdfz", ".xdf.gz")]):  # XDF
                rows, disabled = [], []
                for idx, s in enumerate(resolve_streams(fname)):
                    rows.append([s["stream_id"], s["name"], s["type"], s["channel_count"],
                                 s["channel_format"], s["nominal_srate"]])
                    is_marker = (s["nominal_srate"] == 0 or s["channel_format"] == "string")
                    if is_marker:  # disable marker streams
                        disabled.append(idx)

                enabled = list(set(range(len(rows))) - set(disabled))
                if enabled:
                    selected = enabled[0]
                else:
                    selected = None
                dialog = XDFStreamsDialog(
                    self,
                    rows,
                    fname=fname,
                    selected=selected,
                    disabled=disabled,
                )
                if dialog.exec():
                    rows = [r.row() for r in dialog.view.selectionModel().selectedRows()]
                    stream_ids = [dialog.view.item(r, 0).value() for r in rows]
                    prefix_markers = dialog.prefix_markers
                    kwargs = {}
                    if prefix_markers:
                        kwargs["prefix_markers"] = prefix_markers
                    if dialog.resample.isChecked():
                        kwargs["fs_new"] = float(dialog.fs_new.value())
                    self.model.load(fname, stream_ids=stream_ids, **kwargs)
            elif ext.lower() == ".mat":
                dialog = MatDialog(self, Path(fname).name, parse_mat(fname))
                if dialog.exec():
                    self.model.load(
                        fname,
                        variable=dialog.name,
                        fs=dialog.fs,
                        transpose=dialog.transpose,
                    )
            else:  # all other file formats
                try:
                    self.model.load(fname)
                except FileNotFoundError as e:
                    QMessageBox.critical(self, "File not found", str(e))
                except ValueError as e:
                    QMessageBox.critical(self, "Unknown file type", str(e))

    def open_file(self, f, text, ffilter="*"):
        """Open file."""
        fname = QFileDialog.getOpenFileName(self, text, filter=ffilter)[0]
        if fname:
            f(fname)

    def xdf_chunks(self):
        """Inspect XDF chunks."""
        fname = QFileDialog.getOpenFileName(self, "Select XDF file",
                                            filter="*.xdf *.xdfz *.xdf.gz")[0]
        if fname:
            chunks = list_chunks(fname)
            dialog = XDFChunksDialog(self, chunks, fname)
            dialog.exec()

    def export_file(self, f, text, ffilter="*"):
        """Export to file."""
        fname = QFileDialog.getSaveFileName(self, text)[0]
        if fname:
            exts = [ext.replace("*", "") for ext in ffilter.split()]
            for ext in exts:
                if fname.endswith(ext):
                    return f(fname)
            return f(fname + exts[0])

    def import_file(self, f, text, ffilter="*"):
        """Import file."""
        fname = QFileDialog.getOpenFileName(self, text, filter=ffilter)[0]
        if fname:
            try:
                f(fname)
            except LabelsNotFoundError as e:
                QMessageBox.critical(self, "Channel labels not found", str(e))
            except InvalidAnnotationsError as e:
                QMessageBox.critical(self, "Invalid annotations", str(e))

    def close_all(self):
        """Close all currently open data sets."""
        msg = QMessageBox.question(self, "Close all data sets", "Close all data sets?")
        if msg == QMessageBox.Yes:
            while len(self.model) > 0:
                self.model.remove_data()

    def xdf_metadata(self, fname=None):
        """Show XDF metadata."""
        if fname is None:
            fname = self.model.current["fname"]
        xml = get_xml(fname)
        dialog = XDFMetadataDialog(self, xml)
        dialog.exec()

    def pick_channels(self):
        """Pick channels in current data set."""
        channels = self.model.current["data"].info["ch_names"]
        types = sorted(set(self.model.current["data"].get_channel_types()))
        dialog = PickChannelsDialog(self, channels, types)
        if dialog.exec():
            if dialog.by_name.isChecked():
                picks = [item.text() for item in dialog.names.selectedItems()]
                if set(channels) == set(picks):
                    return
            else:  # by type
                picks = [item.text() for item in dialog.types.selectedItems()]
                if set(types) == set(picks):
                    return
            self.auto_duplicate()
            self.model.pick_channels(picks)

    def channel_properties(self):
        """Show channel properties dialog."""
        info = self.model.current["data"].info
        dialog = ChannelPropertiesDialog(self, info)
        if dialog.exec():
            dialog.model.sort(0)
            bads = []
            renamed = {}
            types = {}
            for i in range(dialog.model.rowCount()):
                new_label = dialog.model.item(i, 1).data(Qt.DisplayRole)
                old_label = info["ch_names"][i]
                if new_label != old_label:
                    renamed[old_label] = new_label
                new_type = dialog.model.item(i, 2).data(Qt.DisplayRole).lower()
                old_type = channel_type(info, i).lower()
                if new_type != old_type:
                    types[new_label] = new_type
                if dialog.model.item(i, 3).checkState() == Qt.Checked:
                    bads.append(info["ch_names"][i])
            self.model.set_channel_properties(bads, renamed, types)

    def rename_channels(self):
        dialog = RenameChannelsDialog(self, self.model.current["data"].info["ch_names"])
        if dialog.exec():
            self.model.rename_channels(dialog.new_names)

    def set_montage(self):
        """Set montage."""
        montages = natural_sort(mne.channels.get_builtin_montages())
        dialog = MontageDialog(self, montages)
        if dialog.exec():
            name = dialog.montages.selectedItems()[0].data(0)
            montage = mne.channels.make_standard_montage(name)
            ch_names = self.model.current["data"].info["ch_names"]
            # check if at least one channel name matches a name in the montage
            if set(ch_names) & set(montage.ch_names):
                self.model.set_montage(
                    name,
                    match_case=dialog.match_case.isChecked(),
                    match_alias=dialog.match_alias.isChecked(),
                    on_missing="ignore" if dialog.ignore_missing.isChecked() else "raise",
                )
            else:
                QMessageBox.critical(
                    self, "No matching channel names", "Channel names defined in the "
                    "montage do not match any channel name in the data."
                )

    def clear_montage(self):
        self.model.set_montage(None)

    def edit_annotations(self):
        fs = self.model.current["data"].info["sfreq"]
        pos = self.model.current["data"].annotations.onset
        pos = (pos * fs).astype(int).tolist()
        dur = self.model.current["data"].annotations.duration
        dur = (dur * fs).astype(int).tolist()
        desc = self.model.current["data"].annotations.description.tolist()
        dialog = AnnotationsDialog(self, pos, dur, desc)
        if dialog.exec():
            rows = dialog.table.rowCount()
            onset, duration, description = [], [], []
            for i in range(rows):
                data = dialog.table.item(i, 0).data(Qt.DisplayRole)
                onset.append(float(data) / fs)
                data = dialog.table.item(i, 1).data(Qt.DisplayRole)
                duration.append(float(data) / fs)
                data = dialog.table.item(i, 2).data(Qt.DisplayRole)
                description.append(data)
            self.model.set_annotations(onset, duration, description)

    def edit_events(self):
        pos = self.model.current["events"][:, 0].tolist()
        desc = self.model.current["events"][:, 2].tolist()
        dialog = EventsDialog(self, pos, desc, self.model.current["event_mapping"])
        if dialog.exec():
            rows = dialog.event_table.rowCount()
            events = np.zeros((rows, 3), dtype=int)
            for i in range(rows):
                pos = dialog.event_table.item(i, 0).value()
                desc = dialog.event_table.item(i, 1).value()
                events[i] = pos, 0, desc
            self.model.current["event_mapping"] = dict(dialog.event_mapping)
            if self.model.current["dtype"] == "epochs":
                event_id_old = self.model.current["data"].event_id
                event_id_new = {f"{k} ({v})": k for k, v in dialog.event_mapping.items() if k in event_id_old.values()}  # noqa: E501
                self.model.current["data"].event_id = event_id_new
            self.model.set_events(events)

    def crop(self):
        """Crop data."""
        stop = self.model.current["data"].times[-1]
        dialog = CropDialog(self, 0, stop)
        if dialog.exec():
            self.auto_duplicate()
            self.model.crop(max(dialog.start, 0), min(dialog.stop, stop))

    def append_data(self):
        """Concatenate raw data objects to current one."""
        compatibles = self.model.get_compatibles()
        dialog = AppendDialog(self, compatibles)
        if dialog.exec():
            self.auto_duplicate()
            self.model.append_data(dialog.names)

    def plot_data(self):
        """Plot data."""
        # self.bad is needed to update history if bad channels are selected in
        # the interactive plot window (see also self.eventFilter)
        self.bads = self.model.current["data"].info["bads"]
        events = self.model.current["events"]
        nchan = self.model.current["data"].info["nchan"]
        fig = self.model.current["data"].plot(events=events, n_channels=nchan,
                                              title=self.model.current["name"],
                                              scalings="auto", show=False)
        if events is not None:
            hist = f"data.plot(events=events, n_channels={nchan})"
        else:
            hist = f"data.plot(n_channels={nchan})"
        self.model.history.append(hist)
        win = fig.canvas.manager.window
        win.setWindowTitle(self.model.current["name"])
        win.statusBar().hide()  # not necessary since matplotlib 3.3
        win.installEventFilter(self)  # detect if the figure is closed

        # prevent closing the window with the escape key
        try:
            fig._mne_params["close_key"] = None
        except AttributeError:  # does not exist in older MNE versions
            pass

        fig.show()

    def plot_psd(self):
        """Plot power spectral density (PSD)."""
        kwds = {}
        if self.model.current["dtype"] == "raw":
            kwds.update({"average": False, "spatial_colors": False})
        fig = self.model.current["data"].plot_psd(show=False, **kwds)
        if kwds:
            tmp = ", ".join(f"{key}={value}" for key, value in kwds.items())
            hist = f"data.plot_psd({tmp})"
        else:
            hist = "data.plot_psd()"
        self.model.history.append(hist)
        win = fig.canvas.manager.window
        win.setWindowTitle("Power spectral density")
        fig.show()

    def plot_locations(self):
        """Plot current montage."""
        fig = self.model.current["data"].plot_sensors(show_names=True, show=False)
        win = fig.canvas.manager.window
        win.setWindowTitle("Montage")
        win.statusBar().hide()  # not necessary since matplotlib 3.3
        fig.show()

    def plot_ica_components(self):
        self.model.current["ica"].plot_components(inst=self.model.current["data"])

    def plot_ica_sources(self):
        self.model.current["ica"].plot_sources(inst=self.model.current["data"])

    def plot_erds(self):
        """Plot ERDS maps."""
        data = self.model.current["data"]
        t_range = [data.tmin, data.tmax]
        f_range = [1, data.info["sfreq"] / 2]

        dialog = ERDSDialog(self, t_range, f_range)

        if dialog.exec():
            freqs = np.arange(dialog.f1, dialog.f2, dialog.step)
            baseline = [dialog.b1, dialog.b2]
            times = [dialog.t1, dialog.t2]
            alpha = None
            if dialog.significance_mask.isChecked():
                alpha = dialog.alpha.value()

            calc = CalcDialog(
                self,
                "Calculating ERDS maps",
                "Calculating ERDS maps..."
            )

            def callback(x):
                QMetaObject.invokeMethod(calc, "accept", Qt.QueuedConnection)

            pool = mp.Pool(processes=1)
            res = pool.apply_async(
                func=_calc_tfr,
                args=(data, freqs, baseline, times, alpha),
                callback=callback
            )
            pool.close()

            if not calc.exec():
                pool.terminate()
                print("ERDS map calculation aborted.")
            else:
                tfr_and_masks = res.get(timeout=1)
                figs = plot_erds(tfr_and_masks)
                for fig in figs:
                    fig.show()

    def plot_erds_topomaps(self):
        """Plot ERDS topomaps."""
        epochs = self.model.current["data"]
        t_range = [epochs.tmin, epochs.tmax]
        f_range = [1, epochs.info["sfreq"] / 2]

        dialog = ERDSTopomapsDialog(self, t_range, f_range, epochs.event_id)
        if dialog.exec():
            figs = plot_erds_topomaps(
                epochs,
                events=[item.text() for item in dialog.events.selectedItems()],
                freqs=np.arange(dialog.f1, dialog.f2, dialog.step),
                baseline=[dialog.b1, dialog.b2],
                times=[dialog.t1, dialog.t2],
            )
            for fig in figs:
                fig.show()

    def plot_evoked(self):
        """Plot evoked potentials for individual channels."""
        epochs = self.model.current["data"]
        dialog = PlotEvokedDialog(
            self,
            epochs.ch_names,
            epochs.event_id,
            epochs.get_montage(),
        )
        if dialog.exec():
            if dialog.topomaps.isChecked():
                if dialog.topomaps_peaks.isChecked():
                    topomap_times = "peaks"
                elif dialog.topomaps_auto.isChecked():
                    topomap_times = "auto"
                else:
                    topomap_times = [float(t.strip()) for t in dialog.topomaps_timelist.text().split(",")]  # noqa: E501
            else:
                topomap_times = []
            figs = plot_evoked(
                epochs=epochs,
                picks=[item.text() for item in dialog.picks.selectedItems()],
                events=[item.text() for item in dialog.events.selectedItems()],
                gfp=dialog.gfp.isChecked(),
                spatial_colors=dialog.spatial_colors.isChecked(),
                topomap_times=topomap_times,
            )
            for fig in figs:
                fig.show()

    def plot_evoked_comparison(self):
        """Plot evoked potentials averaged over channels."""
        epochs = self.model.current["data"]
        dialog = PlotEvokedComparisonDialog(self, epochs.ch_names, epochs.event_id)
        if dialog.exec():
            figs = plot_evoked_comparison(
                epochs=epochs,
                picks=[item.text() for item in dialog.picks.selectedItems()],
                events=[item.text() for item in dialog.events.selectedItems()],
                average_method=dialog.average_epochs.currentText(),
                combine=dialog.combine_channels.currentText(),
                confidence_intervals=dialog.confidence_intervals.isChecked(),
            )
            for fig in figs:
                fig.show()

    def plot_evoked_topomaps(self):
        """Plot evoked topomaps."""
        epochs = self.model.current["data"]
        dialog = PlotEvokedTopomaps(self, epochs.event_id)
        if dialog.exec():
            if dialog.auto.isChecked():
                times = "auto"
            elif dialog.peaks.isChecked():
                times = "peaks"
            elif dialog.interactive.isChecked():
                times = "interactive"
            else:
                times = [float(t.strip()) for t in dialog.timelist.text().split(",")]

            figs = plot_evoked_topomaps(
                epochs=epochs,
                events=[item.text() for item in dialog.events.selectedItems()],
                average_method=dialog.average_epochs.currentText(),
                times=times,
            )
            for fig in figs:
                fig.show()

    def run_ica(self):
        """Run ICA calculation."""

        methods = ["Infomax"]
        if have["picard"]:
            methods.insert(0, "Picard")
        if have["sklearn"]:
            methods.append("FastICA")

        dialog = RunICADialog(self, self.model.current["data"].info["nchan"], methods)

        if dialog.exec():
            calc = CalcDialog(self, "Calculating ICA", "Calculating ICA.")

            method = dialog.method.currentText().lower()
            exclude_bad_segments = dialog.exclude_bad_segments.isChecked()

            fit_params = {}
            if dialog.extended.isEnabled():
                fit_params["extended"] = dialog.extended.isChecked()
            if dialog.ortho.isEnabled():
                fit_params["ortho"] = dialog.ortho.isChecked()

            ica = mne.preprocessing.ICA(method=method, fit_params=fit_params)
            history = f"ica = mne.preprocessing.ICA(method='{method}'"
            if fit_params:
                history += f", fit_params={fit_params})"
            else:
                history += ")"
            self.model.history.append(history)

            pool = mp.Pool(processes=1)

            def callback(x):
                QMetaObject.invokeMethod(calc, "accept", Qt.QueuedConnection)

            res = pool.apply_async(
                func=ica.fit,
                args=(self.model.current["data"],),
                kwds={"reject_by_annotation": exclude_bad_segments},
                callback=callback
            )
            pool.close()

            if not calc.exec():
                pool.terminate()
                print("ICA calculation aborted...")
            else:
                self.model.current["ica"] = res.get(timeout=1)
                self.model.history.append(f"ica.fit(inst=raw, reject_by_annotation="
                                          f"{exclude_bad_segments})")
                self.data_changed()

    def apply_ica(self):
        """Apply current fitted ICA."""
        self.auto_duplicate()
        self.model.apply_ica()

    def interpolate_bads(self):
        """Interpolate bad channels"""
        dialog = InterpolateBadsDialog(self)
        if dialog.exec():
            duplicated = self.auto_duplicate()
            try:
                self.model.interpolate_bads(dialog.reset_bads, dialog.mode, dialog.origin)
            except ValueError as e:
                if duplicated:  # undo
                    self.model.remove_data()
                    self.model.index -= 1
                    self.data_changed()
                msgbox = ErrorMessageBox(self, "Could not interpolate bad channels", str(e),
                                         traceback.format_exc())
                msgbox.show()

    def filter_data(self):
        """Filter data."""
        dialog = FilterDialog(self)
        if dialog.exec():
            self.auto_duplicate()
            self.model.filter(dialog.low, dialog.high)

    def find_events(self):
        info = self.model.current["data"].info

        # use first stim channel as default in dialog
        default_stim = 0
        for i in range(info["nchan"]):
            if mne.io.pick.channel_type(info, i) == "stim":
                default_stim = i
                break
        dialog = FindEventsDialog(self, info["ch_names"], default_stim)
        if dialog.exec():
            stim_channel = dialog.stimchan.currentText()
            consecutive = dialog.consecutive.currentText().lower()
            if consecutive == "true":
                consecutive = True
            elif consecutive == "false":
                consecutive = False
            initial_event = dialog.initial_event.isChecked()
            uint_cast = dialog.uint_cast.isChecked()
            min_dur = dialog.minduredit.value()
            shortest_event = dialog.shortesteventedit.value()
            self.model.find_events(
                stim_channel=stim_channel,
                consecutive=consecutive,
                initial_event=initial_event,
                uint_cast=uint_cast,
                min_duration=min_dur,
                shortest_event=shortest_event
            )

    def events_from_annotations(self):
        self.model.events_from_annotations()

    def annotations_from_events(self):
        self.model.annotations_from_events()

    def epoch_data(self):
        """Epoch raw data."""
        event_types = np.unique(self.model.current["events"][:, 2]).astype(str).tolist()
        dialog = EpochDialog(self, event_types)
        if dialog.exec():
            tmin = dialog.tmin.value()
            tmax = dialog.tmax.value()

            if dialog.baseline.isChecked():
                baseline = dialog.a.value(), dialog.b.value()
            else:
                baseline = None

            duplicated = self.auto_duplicate()

            try:
                self.model.epoch_data(dialog.selected_events, tmin, tmax, baseline)
            except ValueError as e:
                if duplicated:  # undo
                    self.model.remove_data()
                    self.model.index -= 1
                    self.data_changed()
                msgbox = ErrorMessageBox(self, "Could not create epochs", str(e),
                                         traceback.format_exc())
                msgbox.show()

    def drop_bad_epochs(self):
        """Drop bad epochs."""
        def fields_to_dict(fields):
            res = {}
            for type, value in fields.items():
                if value.text():
                    res[type] = float(value.text())
            return res

        types = sorted(set(self.model.current["data"].get_channel_types()))
        dialog = DropBadEpochsDialog(self, types)
        if dialog.exec():
            reject = None
            flat = None
            if dialog.reject_box.isChecked():
                reject = fields_to_dict(dialog.reject_fields)
            if dialog.flat_box.isChecked():
                flat = fields_to_dict(dialog.flat_fields)
            if reject is None and flat is None:
                return
            self.auto_duplicate()
            self.model.drop_bad_epochs(reject, flat)

    def convert_od(self):
        """Convert to optical density."""
        self.auto_duplicate()
        self.model.convert_od()

    def convert_bl(self):
        """Convert to haemoglobin."""
        self.auto_duplicate()
        self.model.convert_beer_lambert()

    def change_reference(self):
        """Change reference."""
        dialog = ReferenceDialog(self, self.model.current["data"].info["ch_names"])
        if dialog.exec():
            if dialog.add_group.isChecked():
                add = [c.strip() for c in dialog.add_channellist.text().split(",")]
            else:
                add = []
            if dialog.reref_group.isChecked():
                if dialog.reref_average.isChecked():
                    ref = "average"
                else:
                    ref = [c.text() for c in dialog.reref_channellist.selectedItems()]
            else:
                ref = None
            duplicated = self.auto_duplicate()
            try:
                self.model.change_reference(add, ref)
            except ValueError as e:
                if duplicated:  # undo
                    self.model.remove_data()
                    # self.model.index -= 1
                    self.data_changed()
                msgbox = ErrorMessageBox(
                    self,
                    "Error while changing references:",
                    str(e),
                    traceback.format_exc(),
                )
                msgbox.show()

    def show_history(self):
        """Show history."""
        dialog = HistoryDialog(self, "\n".join(self.model.history))
        dialog.exec()

    def show_about(self):
        """Show About dialog."""
        from . import __version__

        msg_box = QMessageBox(self)
        text = (f"<img src='{image_path('mnelab_logo.png')}'><p>MNELAB {__version__}</p>")
        msg_box.setText(text)

        mnelab_url = "github.com/cbrnr/mnelab"
        mne_url = "github.com/mne-tools/mne-python"

        pkgs = []
        for key, value in have.items():
            if value:
                pkgs.append(f"{key}&nbsp;({value})")
            else:
                pkgs.append(f"{key}&nbsp;(not installed)")
        version = ".".join(str(k) for k in version_info[:3])
        text = (f"<nobr><p>This program uses Python {version} and the following packages:"
                f"</p></nobr><p>{', '.join(pkgs)}</p>"
                f"<nobr><p>MNELAB repository: <a href=https://{mnelab_url}>{mnelab_url}</a>"
                f"</p></nobr><nobr><p>MNE repository: "
                f"<a href=https://{mne_url}>{mne_url}</a></p></nobr>"
                f"<p>Licensed under the BSD 3-clause license.</p>"
                f"<p>© MNELAB developers.</p>")
        msg_box.setInformativeText(text)
        msg_box.exec()

    def show_about_qt(self):
        """Show About Qt dialog."""
        QMessageBox.aboutQt(self, "About Qt")

    def show_documentation(self):
        url = QUrl('https://mnelab.readthedocs.io/')
        if not QDesktopServices.openUrl(url):
            QMessageBox.warning(self, 'Open Url', 'Could not open url')

    def auto_duplicate(self):
        """Automatically duplicate current data set.

        If the current data set is stored in a file (i.e. was loaded directly from a file),
        a new data set is automatically created. If the current data set is not stored in a
        file (i.e. was created by operations in MNELAB), a dialog box asks the user if the
        current data set should be overwritten or duplicated.

        Returns
        -------
        duplicated : bool
            True if the current data set was automatically duplicated, False if the current
            data set was overwritten.
        """
        # if current data is stored in a file create a new data set
        if self.model.current["fname"]:
            self.model.duplicate_data()
            return True
        # otherwise ask the user
        msg = QMessageBox(self)
        msg.setWindowFlags(Qt.Dialog | Qt.WindowTitleHint | Qt.CustomizeWindowHint)
        msg.setWindowTitle("Modify data set")
        msg.setText("You are about to modify the current data set. How do you want to proceed?")  # noqa: E501
        create_button = msg.addButton("Create new data set", QMessageBox.AcceptRole)
        overwrite_button = msg.addButton("Overwrite current data set", QMessageBox.RejectRole)  # noqa: E501
        msg.setDefaultButton(create_button)
        msg.setEscapeButton(create_button)
        msg.exec()
        if msg.clickedButton() == overwrite_button:
            return False
        else:
            self.model.duplicate_data()
            return True

    def _add_recent(self, fname):
        """Add a file to recent file list.

        Parameters
        ----------
        fname : str
            File name.
        """
        if fname in self.recent:  # avoid duplicates
            self.recent.remove(fname)
        self.recent.insert(0, fname)
        self.recent = self.recent[:read_settings("max_recent")]  # prune list
        write_settings(recent=self.recent)
        if not self.recent_menu.isEnabled():
            self.recent_menu.setEnabled(True)

    def _remove_recent(self, fname):
        """Remove file from recent file list.

        Parameters
        ----------
        fname : str
            File name.
        """
        if fname in self.recent:
            self.recent.remove(fname)
            write_settings(recent=self.recent)
            if not self.recent:
                self.recent_menu.setEnabled(False)

    @Slot(QModelIndex)
    def _update_data(self, selected):
        """Update index and information based on the state of the sidebar.

        Parameters
        ----------
        selected : QModelIndex
            Index of the selected row.
        """
        if selected.row() != self.model.index:
            self.model.index = selected.row()
            self.data_changed()
            self.model.history.append(f"data = datasets[{self.model.index}]")

    @Slot()
    def _update_recent_menu(self):
        self.recent_menu.clear()
        for recent in self.recent:
            self.recent_menu.addAction(recent)

    @Slot(QAction)
    def _load_recent(self, action):
        self.open_data(fname=action.text())

    @Slot()
    def _toggle_toolbar(self):
        if self.toolbar.isHidden():
            self.toolbar.show()
        else:
            self.toolbar.hide()
        write_settings(toolbar=not self.toolbar.isHidden())

    @Slot()
    def _toggle_statusbar(self):
        if self.statusBar().isHidden():
            self.statusBar().show()
        else:
            self.statusBar().hide()
        write_settings(statusbar=not self.statusBar().isHidden())

    @Slot(QDropEvent)
    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    @Slot(QDropEvent)
    def dropEvent(self, event):
        mime = event.mimeData()
        if mime.hasUrls():
            urls = mime.urls()
            for url in urls:
                try:
                    self.open_data(url.toLocalFile())
                except FileNotFoundError as e:
                    QMessageBox.critical(self, "File not found", str(e))

    @Slot(QEvent)
    def closeEvent(self, event):
        """Close application.

        Parameters
        ----------
        event : QEvent
            Close event.
        """
        write_settings(size=self.size(), pos=self.pos())
        if self.model.history:
            print("\n# Command History\n")
            print("\n".join(self.model.history))
        QApplication.quit()

    def eventFilter(self, source, event):
        # currently the only source is the raw plot window
        if event.type() == QEvent.Close:
            self.data_changed()
            bads = self.model.current["data"].info["bads"]
            if self.bads != bads:
                self.model.history.append(f'data.info["bads"] = {bads}')
        return QObject.eventFilter(self, source, event)

    def event(self, ev):
        """Catch system events."""
        if ev.type() == QEvent.PaletteChange:  # detect theme switches
            style = interface_style()  # light or dark
            if style is not None:
                QIcon.setThemeName(style)
            else:
                QIcon.setThemeName("light")  # fallback
        return super().event(ev)



# © MNELAB developers
#
# License: BSD (3-clause)

import keyword

from PySide6.QtCore import QRegularExpression, Qt
from PySide6.QtGui import QFont, QSyntaxHighlighter, QTextCharFormat


class PythonHighlighter(QSyntaxHighlighter):
    def __init__(self, document):
        super().__init__(document)

        self.rules = []

        # keywords
        f = QTextCharFormat()
        f.setFontWeight(QFont.Bold)
        f.setForeground(Qt.darkBlue)
        for kw in keyword.kwlist:
            self.rules.append((QRegularExpression(rf"\b{kw}\b"), f))

        # numerals
        f = QTextCharFormat()
        f.setForeground(Qt.blue)
        self.rules.append((QRegularExpression("[0-9]+"), f))

        # strings
        f = QTextCharFormat()
        f.setForeground(Qt.darkCyan)
        self.rules.append((QRegularExpression('"[^"]*"'), f))
        self.rules.append((QRegularExpression("'[^']*'"), f))

    def highlightBlock(self, text):
        for rule in self.rules:
            matches = rule[0].globalMatch(text)
            while matches.hasNext():
                match = matches.next()
                start, length = match.capturedStart(), match.capturedLength()
                self.setFormat(start, length, rule[1])



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QListWidget,
    QRadioButton,
    QVBoxLayout,
)


class ReferenceDialog(QDialog):
    def __init__(self, parent, available_channels):
        super().__init__(parent)
        self.setWindowTitle("Change reference")
        vbox = QVBoxLayout(self)

        self.add_group = QGroupBox("Add reference (all zero)")
        self.reref_group = QGroupBox("Re-reference to existing channel(s)")

        vbox.addWidget(self.add_group)
        vbox.addWidget(self.reref_group)

        self.add_group.setCheckable(True)
        self.reref_group.setCheckable(True)

        self.add_group.setChecked(False)

        self.add_channellist = QLineEdit()
        add_grid = QGridLayout()
        add_grid.setColumnStretch(0, 2)
        add_grid.setColumnStretch(1, 3)
        add_grid.addWidget(QLabel("Channel(s):"), 0, 0)
        add_grid.addWidget(self.add_channellist, 0, 1)
        self.add_group.setLayout(add_grid)

        self.reref_average = QRadioButton("Average")
        self.reref_channels = QRadioButton("Channel(s):")
        self.reref_channellist = QListWidget()
        self.reref_channellist.insertItems(0, available_channels)
        self.reref_average.setChecked(True)
        self.reref_channellist.setEnabled(False)
        self.reref_channellist.setSelectionMode(QListWidget.ExtendedSelection)
        reref_grid = QGridLayout()
        reref_grid.setColumnStretch(0, 2)
        reref_grid.setColumnStretch(1, 3)
        reref_grid.addWidget(self.reref_average, 0, 0)
        reref_grid.addWidget(self.reref_channels, 1, 0, alignment=Qt.AlignTop)
        reref_grid.addWidget(self.reref_channellist, 1, 1)

        self.reref_group.setLayout(reref_grid)

        self.reref_channels.toggled.connect(self.toggle_reref_channellist)
        self.add_group.toggled.connect(self.toggle_ok)
        self.add_channellist.textChanged.connect(self.toggle_ok)
        self.reref_group.toggled.connect(self.toggle_ok)
        self.reref_channels.toggled.connect(self.toggle_ok)
        self.reref_channellist.itemSelectionChanged.connect(self.toggle_ok)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    def toggle_reref_channellist(self):
        self.reref_channellist.setEnabled(self.reref_channels.isChecked())

    def toggle_ok(self):
        if self.add_group.isChecked() and not self.add_channellist.text():
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        if not self.add_group.isChecked() and not self.reref_group.isChecked():
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        if self.reref_group.isChecked() and self.reref_channels.isChecked() and not self.reref_channellist.selectedItems():  # noqa: E501
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)



# © MNELAB developers
#
# License: BSD (3-clause)

import numpy as np
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
)


def populate_tree(parent, nodes):
    if isinstance(nodes, list):  # struct (available as list)
        for i, v in enumerate(nodes):
            item = QTreeWidgetItem(parent)
            item.setText(0, f"[{i}]")
            item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
            populate_tree(item, v)
    else:  # dict containing variable/value pairs
        for k, v in nodes.items():
            item = QTreeWidgetItem(parent)
            item.setText(0, k)
            if isinstance(v, dict):
                item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
                populate_tree(item, v)
            elif isinstance(v, list):
                item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
                populate_tree(item, v)
            else:
                item.setText(1, type(v).__name__)
                if isinstance(v, np.ndarray):
                    item.setText(1, f"{type(v).__name__} ({v.dtype.name})")  # add dtype
                    item.setText(2, " × ".join(map(str, v.shape)))
                    if v.ndim > 2:  # arrays with more than two dimensions cannot be loaded
                        item.setFlags(Qt.NoItemFlags)
                    elif v.dtype not in (np.float32, np.float64):  # must be numeric types
                        item.setFlags(Qt.NoItemFlags)
                else:
                    item.setFlags(Qt.NoItemFlags)
                    item.setText(3, repr(v))


class MatDialog(QDialog):
    def __init__(self, parent, fname, nodes):
        super().__init__(parent)
        self.setWindowTitle("Select variable")

        self.tree = QTreeWidget()
        self.tree.setColumnCount(4)
        self.tree.setHeaderLabels(["Name", "Type", "Shape", "Value"])
        self.tree.setColumnWidth(0, 200)
        self.tree.setColumnWidth(1, 125)
        self.tree.setColumnWidth(2, 125)

        self.root = QTreeWidgetItem(self.tree)
        self.root.setText(0, fname)
        self.root.setFlags(self.root.flags() & ~Qt.ItemIsSelectable)

        populate_tree(self.root, nodes)

        self.tree.expandAll()

        vbox = QVBoxLayout(self)
        vbox.addWidget(self.tree)
        hbox = QHBoxLayout()
        self._fs = QDoubleSpinBox()
        self._fs.setMaximum(999999)
        self._fs.setValue(250)
        self._fs.setDecimals(2)
        self._fs.setSuffix(" Hz")
        hbox.addWidget(QLabel("Sampling frequency:"))
        hbox.addWidget(self._fs)
        hbox.addStretch()
        self._transpose = QCheckBox("Transpose")
        self._transpose.setChecked(False)
        hbox.addWidget(self._transpose)
        vbox.addLayout(hbox)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        self.tree.itemSelectionChanged.connect(self.toggle)
        self.toggle()
        self.resize(650, 550)

    @property
    def name(self):
        item = self.tree.selectedItems()[0]
        text = item.text(0)
        parent = item.parent()
        while parent is not self.root:
            text = f"{parent.text(0)}.{text}"
            parent = parent.parent()
        return text

    @property
    def fs(self):
        return self._fs.value()

    @property
    def transpose(self):
        if not self._transpose.isEnabled():
            return False
        return self._transpose.isChecked()

    @Slot()
    def toggle(self):
        """Toggle OK and Transpose buttons."""
        self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
        if (items := self.tree.selectedItems()):
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)
            shape = [int(dim) for dim in items[0].text(2).split(" × ")]
            if len(shape) == 1:
                self._transpose.setEnabled(False)
            else:
                self._transpose.setEnabled(True)
                if shape[0] > shape[1]:  # transpose if there are more rows than columns
                    self._transpose.setChecked(True)
                else:
                    self._transpose.setChecked(False)



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QGridLayout,
    QLineEdit,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)


class RenameChannelsDialog(QDialog):
    def __init__(self, parent, channels):
        super().__init__(parent)
        self.setWindowTitle("Rename channels")

        self.old_names = channels

        self.method = QComboBox()
        self.method.addItems(["Strip characters", "Delete characters"])
        self.method.setCurrentIndex(0)

        self.strip_chars = QLineEdit()
        self.slice_num = QDoubleSpinBox()
        self.slice_num.setMinimum(0)
        self.slice_num.setDecimals(0)

        self.where = QComboBox()
        self.where.addItems(["from beginning", "from end"])
        self.where.setCurrentIndex(0)

        option_grid = QGridLayout()
        option_grid.setColumnStretch(0, 1)
        option_grid.setColumnStretch(1, 1)
        option_grid.setColumnStretch(2, 1)
        option_grid.addWidget(self.method, 0, 0)
        option_grid.addWidget(self.strip_chars, 0, 1)
        option_grid.addWidget(self.slice_num, 0, 1)
        option_grid.addWidget(self.where, 0, 2)

        self.preview = QTableWidget(len(channels), 2)
        self.preview.setHorizontalHeaderLabels(["Before", "After"])
        self.preview.horizontalHeader().setStretchLastSection(True)
        self.preview.verticalHeader().setVisible(False)
        self.preview.setShowGrid(False)
        self.preview.setSelectionBehavior(QTableWidget.SelectRows)
        self.preview.setColumnWidth(0, 190)
        self.preview.setColumnWidth(1, 190)
        self.preview.setEditTriggers(QTableWidget.NoEditTriggers)
        self.preview.setSelectionMode(QTableWidget.NoSelection)
        self.preview.setFocusPolicy(Qt.NoFocus)

        vbox = QVBoxLayout(self)
        vbox.addLayout(option_grid)
        vbox.addSpacing(10)
        vbox.addWidget(self.preview)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        self.method.currentTextChanged.connect(self.toggle_input)
        self.method.currentTextChanged.connect(self.update_preview)
        self.where.currentTextChanged.connect(self.update_preview)
        self.strip_chars.textEdited.connect(self.update_preview)
        self.slice_num.valueChanged.connect(self.update_preview)

        self.toggle_input()
        self.update_preview()
        self.setFixedSize(450, 450)

    @Slot()
    def toggle_input(self):
        if self.method.currentText() == "Strip characters":
            self.strip_chars.setVisible(True)
            self.slice_num.setVisible(False)
        elif self.method.currentText() == "Delete characters":
            self.strip_chars.setVisible(False)
            self.slice_num.setVisible(True)

    @Slot()
    def update_preview(self):
        if self.method.currentText() == "Strip characters":
            if self.where.currentText() == "from beginning":
                new_names = [n.lstrip(self.strip_chars.text()) for n in self.old_names]
            else:
                new_names = [n.rstrip(self.strip_chars.text()) for n in self.old_names]

        else:
            if self.where.currentText() == "from beginning":
                new_names = [n[int(self.slice_num.value()):] for n in self.old_names]
            else:
                if self.slice_num.value() > 0:
                    new_names = [n[:-int(self.slice_num.value())] for n in self.old_names]
                else:
                    new_names = self.old_names[:]

        self.new_names = new_names
        for row, (old, new) in enumerate(zip(self.old_names, self.new_names)):
            self.preview.setItem(row, 0, QTableWidgetItem(old))
            self.preview.setItem(row, 1, QTableWidgetItem(new))



# © MNELAB developers
#
# License: BSD (3-clause)

import multiprocessing as mp
import sys
from pathlib import Path

import matplotlib
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication

from .mainwindow import MainWindow
from .model import Model

__version__ = "0.8.4"


def main():
    mp.set_start_method("spawn", force=True)  # required for Linux
    app_name = "MNELAB"
    if sys.platform.startswith("darwin"):
        # set bundle name on macOS (app name shown in the menu bar)
        # this must be done before the app is created
        from Foundation import NSBundle
        bundle = NSBundle.mainBundle()
        info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
        info["CFBundleName"] = app_name

    matplotlib.use("QtAgg")
    app = QApplication(sys.argv)
    app.setApplicationName(app_name)
    app.setOrganizationName("cbrnr")
    if sys.platform.startswith("darwin"):
        app.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus, True)
        app.setWindowIcon(QIcon(f"{Path(__file__).parent}/icons/mnelab-logo-macos.svg"))
    else:
        app.setWindowIcon(QIcon(f"{Path(__file__).parent}/icons/mnelab-logo.svg"))
    app.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)
    model = Model()
    model.view = MainWindow(model)
    if len(sys.argv) > 1:  # open files from command line arguments
        for f in sys.argv[1:]:
            model.load(f)
    model.view.show()
    sys.exit(app.exec())



# © MNELAB developers
#
# License: BSD (3-clause)

from collections import defaultdict

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

from .utils import FloatTableWidgetItem, IntTableWidgetItem


class XDFStreamsDialog(QDialog):
    def __init__(self, parent, rows, fname, selected=None, disabled=None):
        super().__init__(parent)
        self.setWindowTitle("Select XDF Stream")
        self.fname = fname

        self.view = QTableWidget(len(rows), 6)
        for i, row in enumerate(rows):
            self.view.setItem(i, 0, IntTableWidgetItem(row[0]))
            self.view.setItem(i, 1, QTableWidgetItem(row[1]))
            self.view.setItem(i, 2, QTableWidgetItem(row[2]))
            self.view.setItem(i, 3, IntTableWidgetItem(row[3]))
            self.view.setItem(i, 4, QTableWidgetItem(row[4]))
            self.view.setItem(i, 5, FloatTableWidgetItem(row[5]))
            if i in disabled:
                for col in range(6):
                    self.view.item(i, col).setFlags(Qt.NoItemFlags)
        self.view.setHorizontalHeaderLabels([
            "ID",
            "Name",
            "Type",
            "Channels",
            "Format",
            "Sampling Rate",
        ])

        self.view.setEditTriggers(QTableWidget.NoEditTriggers)
        self.view.setSelectionBehavior(QTableWidget.SelectRows)
        self.view.verticalHeader().setVisible(False)
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.setShowGrid(False)
        if selected is not None:
            self.view.selectRow(selected)
        self.view.setSortingEnabled(True)
        self.view.sortByColumn(0, Qt.AscendingOrder)

        self.view.itemSelectionChanged.connect(self.toggle_buttons)

        self.resample = QCheckBox()
        self.resample_label = QLabel("Resample to:")
        self.fs_new = QDoubleSpinBox()
        self.fs_new.setRange(1,  max(r[5] for r in rows))
        self.fs_new.setValue(1)
        self.fs_new.setDecimals(1)
        self.fs_new.setSuffix(" Hz")

        self._prefix_markers = QCheckBox("Prefix markers with stream ID")
        self._prefix_markers.setChecked(False)
        if len(disabled) < 2:
            self._prefix_markers.setEnabled(False)

        hbox1 = QHBoxLayout()
        hbox1.addWidget(self.resample)
        hbox1.addWidget(self.resample_label)
        hbox1.addWidget(self.fs_new)
        hbox1.addStretch()
        hbox1.addWidget(self._prefix_markers)
        hbox1.addStretch()

        vbox = QVBoxLayout(self)
        vbox.addWidget(self.view)
        vbox.addLayout(hbox1)

        hbox2 = QHBoxLayout()
        self.details_button = QPushButton("Details")
        self.details_button.clicked.connect(self.details)
        hbox2.addWidget(self.details_button)
        hbox2.addStretch()
        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        hbox2.addWidget(self.buttonbox)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        vbox.addLayout(hbox2)

        self.toggle_buttons()
        self.resize(775, 650)
        self.view.setColumnWidth(0, 90)
        self.view.setColumnWidth(1, 200)
        self.view.setColumnWidth(2, 140)

    @property
    def prefix_markers(self):
        return self._prefix_markers.isChecked()

    def details(self):
        self.parent().xdf_metadata(self.fname)

    @Slot()
    def toggle_buttons(self):
        # toggle the resample selection
        if len(self.view.selectionModel().selectedRows()) > 1:
            self.resample.setChecked(True)
            self.resample.setEnabled(False)
        else:
            self.resample.setChecked(False)
            self.resample.setEnabled(True)

        # if there is no selection disable OK
        if not self.view.selectionModel().selectedRows():
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
        else:
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)
            # suggest the most common sampling rate (with the most channels)
            channel_counts = []
            sampling_rates = []
            row_indices = {r.row() for r in self.view.selectedIndexes()}
            for row in row_indices:
                channel_counts.append(self.view.item(row, 3).value())
                sampling_rates.append(self.view.item(row, 5).value())

            counts = defaultdict(int)
            for channel_count, sampling_rate in zip(channel_counts, sampling_rates):
                if sampling_rate != 0:
                    counts[sampling_rate] += channel_count

            suggested_fs = max(counts, key=counts.get)
            self.fs_new.setValue(suggested_fs)



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QListWidget,
    QRadioButton,
)

from .utils import select_all


class PickChannelsDialog(QDialog):
    def __init__(self, parent, channels, types):
        super().__init__(parent)
        self.setWindowTitle("Pick channels")

        grid = QGridLayout(self)

        self.by_name = QRadioButton("By name:")
        grid.addWidget(self.by_name, 0, 0, Qt.AlignTop)
        self.names = QListWidget()
        self.names.insertItems(0, channels)
        self.names.setSelectionMode(QListWidget.ExtendedSelection)
        select_all(self.names)
        grid.addWidget(self.names, 0, 1)
        self.by_name.setChecked(True)

        self.by_type = QRadioButton("By type:")
        grid.addWidget(self.by_type, 1, 0, Qt.AlignTop)
        self.types = QListWidget()
        self.types.insertItems(0, types)
        self.types.setSelectionMode(QListWidget.ExtendedSelection)
        self.types.setMaximumHeight(self.types.sizeHintForRow(0) * 5.5)
        select_all(self.types)
        grid.addWidget(self.types, 1, 1)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        grid.addWidget(self.buttonbox, 2, 0, 1, -1)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        self.types.itemSelectionChanged.connect(self.toggle_buttons)
        self.names.itemSelectionChanged.connect(self.toggle_buttons)
        self.by_name.toggled.connect(self.toggle_buttons)
        self.toggle_buttons()  # initialize OK button state

        self.by_name.toggled.connect(self.toggle_lists)
        self.by_type.toggled.connect(self.toggle_lists)
        self.toggle_lists()

    @Slot()
    def toggle_buttons(self):
        """Toggle OK button."""
        self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(False)
        if (self.by_name.isChecked() and self.names.selectedItems()
                or self.by_type.isChecked() and self.types.selectedItems()):
            self.buttonbox.button(QDialogButtonBox.Ok).setEnabled(True)

    @Slot()
    def toggle_lists(self):
        self.names.setEnabled(self.by_name.isChecked())
        self.types.setEnabled(not self.by_name.isChecked())



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtCore import QPoint, QSettings, QSize, Slot
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
)

_DEFAULTS = {
    "max_recent": 6,
    "recent": [],
    "toolbar": True,
    "statusbar": True,
    "size": QSize(700, 500),
    "pos": QPoint(100, 100),
}


def _get_value(key):
    # The third argument to QSettings().value is a typecast which is applied to the
    # retrieved value. Since booleans are stored as "true" and "false" (at least on Win10),
    # we need that typecast. However, lists would get replaced with an empty list for some
    # reason, so here's a special case:
    if isinstance(_DEFAULTS[key], list):
        return QSettings().value(key, defaultValue=_DEFAULTS[key])
    return QSettings().value(key, defaultValue=_DEFAULTS[key], type=type(_DEFAULTS[key]))


def read_settings(key=None):
    """
    Read application settings.

    Parameters
    ----------
    key : str, optional
        Setting key to read, by default `None`.

    Returns
    -------
    str | dict
        If `key` is given, the corresponding value. If key is `None`, return all settings in
        a dictionary.
    """
    if key is not None:
        return _get_value(key)
    return {key: _get_value(key) for key in _DEFAULTS}


def write_settings(**kwargs):
    """Write application settings."""
    for key, value in kwargs.items():
        QSettings().setValue(key, value)


class SettingsDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("MNELAB settings")

        grid = QGridLayout(self)

        grid.addWidget(QLabel("Maximum recent files:"), 0, 0)
        self.max_recent = QDoubleSpinBox()
        self.max_recent.setRange(5, 25)
        self.max_recent.setValue(_get_value("max_recent"))
        self.max_recent.setDecimals(0)
        grid.addWidget(self.max_recent, 0, 1)

        self.reset_to_defaults = QPushButton("Reset to defaults")
        self.reset_to_defaults.clicked.connect(self.reset_settings)
        grid.addWidget(self.reset_to_defaults, 1, 0)

        self.reset_to_defaults = QPushButton("Reset window")
        self.reset_to_defaults.clicked.connect(self.reset_window)
        grid.addWidget(self.reset_to_defaults, 2, 0)

        hbox = QHBoxLayout()
        self.buttonbox = QDialogButtonBox(
            QDialogButtonBox.Apply
            | QDialogButtonBox.Ok
            | QDialogButtonBox.Cancel
        )
        hbox.addWidget(self.buttonbox)
        grid.addLayout(hbox, 3, 0, 1, 2)
        self.buttonbox.button(QDialogButtonBox.Apply).clicked.connect(self.apply_settings)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.accepted.connect(self.apply_settings)
        self.buttonbox.rejected.connect(self.reject)

    @Slot()
    def apply_settings(self):
        QSettings().setValue("max_recent", int(self.max_recent.text()))
        self.parent().recent = self.parent().recent[:_get_value("max_recent")]
        QSettings().setValue("recent", self.parent().recent)

    @Slot()
    def reset_settings(self):
        self.max_recent.setValue(_DEFAULTS["max_recent"])

    @Slot()
    def reset_window(self):
        self.parent().resize(_DEFAULTS["size"])
        self.parent().move(_DEFAULTS["pos"])



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
)


class FilterDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Filter data")
        vbox = QVBoxLayout(self)
        grid = QGridLayout()
        grid.addWidget(QLabel("Low cutoff frequency (Hz):"), 0, 0)
        self.lowedit = QLineEdit()
        grid.addWidget(self.lowedit, 0, 1)
        grid.addWidget(QLabel("High cutoff frequency (Hz):"), 1, 0)
        self.highedit = QLineEdit()
        grid.addWidget(self.highedit, 1, 1)
        vbox.addLayout(grid)
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        vbox.addWidget(buttonbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        vbox.setSizeConstraint(QVBoxLayout.SetFixedSize)

    @property
    def low(self):
        low = self.lowedit.text()
        return float(low) if low else None

    @property
    def high(self):
        high = self.highedit.text()
        return float(high) if high else None



# © MNELAB developers
#
# License: BSD (3-clause)

from PySide6.QtWidgets import QDialog, QDialogButtonBox, QLabel, QVBoxLayout


class CalcDialog(QDialog):
    def __init__(self, parent, title, message):
        super().__init__(parent)
        self.setWindowTitle(title)
        vbox = QVBoxLayout(self)
        label = QLabel(message)
        button = QDialogButtonBox(QDialogButtonBox.Cancel)
        button.rejected.connect(self.close)
        vbox.addWidget(label)
        vbox.addWidget(button)
        self.resize(300, 100)



# © MNELAB developers
#
# License: BSD (3-clause)

from collections import defaultdict

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

from .utils import IntTableWidgetItem


class EventsDialog(QDialog):
    def __init__(self, parent, pos, desc, event_mapping):
        super().__init__(parent)
        self.setWindowTitle("Edit Events")

        self.event_table = QTableWidget(len(pos), 2)

        for row, (p, d) in enumerate(zip(pos, desc)):
            self.event_table.setItem(row, 0, IntTableWidgetItem(p))
            self.event_table.setItem(row, 1, IntTableWidgetItem(d))

        self.event_table.setHorizontalHeaderLabels(["Position", "Type"])
        self.event_table.horizontalHeader().setStretchLastSection(True)
        self.event_table.verticalHeader().setVisible(False)
        self.event_table.setShowGrid(False)
        self.event_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.event_table.setSortingEnabled(True)
        self.event_table.sortByColumn(0, Qt.AscendingOrder)

        self.event_mapping = defaultdict(str, event_mapping)  # make copy

        vbox = QVBoxLayout(self)
        vbox.addWidget(self.event_table)
        hbox = QHBoxLayout()
        self.add_button = QPushButton("+")
        self.remove_button = QPushButton("-")
        self.mapping_button = QPushButton("Mapping...")
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        hbox.addWidget(self.add_button)
        hbox.addWidget(self.remove_button)
        hbox.addWidget(self.mapping_button)
        hbox.addStretch()
        hbox.addWidget(buttonbox)
        vbox.addLayout(hbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        self.event_table.itemSelectionChanged.connect(self.toggle_buttons)
        self.remove_button.clicked.connect(self.remove_event)
        self.add_button.clicked.connect(self.add_event)
        self.remove_button.clicked.connect(self.toggle_buttons)
        self.add_button.clicked.connect(self.toggle_buttons)
        self.mapping_button.clicked.connect(self.open_mapping_dialog)
        self.toggle_buttons()
        self.setMinimumSize(500, 500)
        self.resize(500, 500)

    @property
    def unique_events(self):
        _unique_events = set()
        for i in range(self.event_table.rowCount()):
            if item := self.event_table.item(i, 1):
                _unique_events.add(int(item.value()))
        return _unique_events

    @Slot()
    def open_mapping_dialog(self):
        dialog = EventMappingDialog(self)
        if dialog.exec():
            self.event_mapping = dialog.event_mapping

    @Slot()
    def toggle_buttons(self):
        """Toggle + and - buttons."""
        n_items = len(self.event_table.selectedItems())
        if self.event_table.rowCount() == 0:  # no events available
            self.add_button.setEnabled(True)
            self.remove_button.setEnabled(False)
            self.mapping_button.setEnabled(False)
        elif n_items == 2:  # one row (2 items) selected
            self.add_button.setEnabled(True)
            self.remove_button.setEnabled(True)
            self.mapping_button.setEnabled(True)
        elif n_items > 2:  # more than one row selected
            self.add_button.setEnabled(False)
            self.remove_button.setEnabled(True)
            self.mapping_button.setEnabled(True)
        else:  # no rows selected
            self.add_button.setEnabled(False)
            self.remove_button.setEnabled(False)
            self.mapping_button.setEnabled(True)

    def add_event(self):
        if self.event_table.selectedIndexes():
            current_row = self.event_table.selectedIndexes()[0].row()
            pos = int(self.event_table.item(current_row, 0).data(Qt.DisplayRole))
        else:
            current_row = 0
            pos = 0
        self.event_table.setSortingEnabled(False)
        self.event_table.insertRow(current_row)
        self.event_table.setItem(current_row, 0, IntTableWidgetItem(pos))
        self.event_table.setItem(current_row, 1, IntTableWidgetItem(0))
        self.event_table.setSortingEnabled(True)

    def remove_event(self):
        rows = {index.row() for index in self.event_table.selectedIndexes()}
        self.event_table.clearSelection()
        for row in sorted(rows, reverse=True):
            value = self.event_table.item(row, 1).value()
            self.event_table.removeRow(row)
            if value not in self.unique_events:
                self.event_mapping.pop(value, None)


class EventMappingDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Event Mapping")
        self.event_table = parent.event_table
        self.unique_events = parent.unique_events
        self.event_mapping = defaultdict(str, parent.event_mapping)  # make copy

        self.mapping_table = QTableWidget(0, 2)
        self.mapping_table.setHorizontalHeaderLabels(["Type", "Label"])
        self.mapping_table.horizontalHeader().setStretchLastSection(True)
        self.mapping_table.verticalHeader().setVisible(False)
        self.mapping_table.setShowGrid(False)
        self.mapping_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.fill_mapping_table()
        self.clear_button = QPushButton("Clear mapping")

        vbox = QVBoxLayout(self)
        vbox.addWidget(self.mapping_table)
        hbox = QHBoxLayout()
        hbox.addWidget(self.clear_button)
        hbox.addStretch()
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        hbox.addWidget(buttonbox)
        vbox.addLayout(hbox)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)

        self.mapping_table.itemChanged.connect(self.store_mapping)
        self.clear_button.clicked.connect(self.clear_mapping)

        self.mapping_table.setMinimumHeight(150)

    def fill_mapping_table(self):
        self.mapping_table.setRowCount(0)
        for row, id_ in enumerate(sorted(self.unique_events)):
            id_item = IntTableWidgetItem(id_)
            id_item.setFlags(id_item.flags() ^ Qt.ItemIsEditable)
            self.mapping_table.insertRow(row)
            self.mapping_table.setItem(row, 0, id_item)
            self.mapping_table.setItem(row, 1, QTableWidgetItem(self.event_mapping[id_]))

    def store_mapping(self):
        for i in range(self.mapping_table.rowCount()):
            event_id = int(self.mapping_table.item(i, 0).value())
            if event_id not in self.unique_events:
                del self.event_mapping[event_id]
            if self.mapping_table.item(i, 1) is not None:
                self.event_mapping[event_id] = self.mapping_table.item(i, 1).text()

    def clear_mapping(self):
        self.event_mapping.clear()
        self.fill_mapping_table()
